#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "guiutils.h"
#include "guirgbimg.h"
#include "imgview.h"
#include "colorlevelsdlg.h"
#include "config.h"

#include "images/icon_color_levels_32x32.xpm"
#include "images/icon_color_levels_48x48.xpm"
#include "images/icon_ok_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"


typedef struct _Dialog			Dialog;


/*
 *	Dialog:
 */
struct _Dialog {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	GdkGC		*gc;
	gint		freeze_count;
	imgview_struct	*iv;

	GdkCursor	*translate_cur;

	GtkWidget	*histogram_da,
			*min_spin,
			*mid_spin,
			*max_spin;
	GdkPixmap	*histogram_pm;
	guint		histogram_data[(gint)((guint8)-1) + 1];
	guint		histogram_data_max;

	gint		drag_button,
			drag_x,
			drag_y,
			drag_marker;

	guint		apply_to_iv_toid;

	/* Copy of original frame */
	guint8		*orig_img_data;
	gint		cur_frame;

};
#define DIALOG(p)	((Dialog *)(p))


static gint delete_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void set_cb(GtkWidget *widget, gpointer data);
static void cancel_cb(GtkWidget *widget, gpointer data);
static void adj_value_changed_cb(GtkAdjustment *adj, gpointer data);
static gint histogram_event_cb(GtkWidget *widget, GdkEvent *event, gpointer data);
static gint make_histogram_idlecb(gpointer data);
static gint apply_iv_image_tocb(gpointer data);

static void get_marker_values_in_bytes(
	Dialog *d, gint *min_rtn, gint *mid_rtn, gint *max_rtn
);
static void set_marker(
	Dialog *d, gint marker, gint i
);

static void histogram_draw(Dialog *d);

static guint8 color_level_image_iterate(
	const gint v,
	const gint min_displace, const gint mid, const gint max_displace,
	const gint min_to_mid, const gint mid_to_max,
	const gint max_minus_mid
);
static void color_level_image(
	const guint8 *src, guint8 *tar,
	const gint width, const gint height,
	const gint bpp, const gint bpl,
	gint min, gint mid, gint max,
	gboolean interruptable
);
static void apply_iv_image(Dialog *d, gboolean interruptable);
static void queue_apply_iv_image(Dialog *d);
static void restore_iv_image(Dialog *d);

static void make_histogram(
	guint *histogram_data,		/* 256 guints */
	guint *histogram_data_max,
	const guint8 *buf,
	const gint width, const gint height,
	const gint bpp, const gint bpl
);
static void build_std_dialog(
	gint width, gint height, const gchar *title,
	GtkWidget *ref_toplevel,
	GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
	GtkAccelGroup **accelgrp_rtn
);

gboolean iv_color_levels(
	GtkWidget *ref_toplevel,
	imgview_struct *iv
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


static gboolean		got_response;


/*
 *	Toplevel GtkWindow "delete_event" signal callback.
 */
static gint delete_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	got_response = FALSE;
	gtk_main_quit();
	return(TRUE);
}

/*
 *	Set callback.
 */
static void set_cb(GtkWidget *widget, gpointer data)
{
	got_response = TRUE;
	gtk_main_quit();
}

/*
 *	Cancel callback.
 */
static void cancel_cb(GtkWidget *widget, gpointer data)
{
	got_response = FALSE;
	gtk_main_quit();
}

/*
 *	GtkAdjustment "value_changed" signal callback.
 */
static void adj_value_changed_cb(GtkAdjustment *adj, gpointer data)
{
	GtkWidget *w;
	Dialog *d = DIALOG(data);
	if(d == NULL)
	    return;

	if(d->freeze_count > 0)
	    return;

	d->freeze_count++;

	w = d->histogram_da;
	if(w != NULL)
	    gtk_widget_queue_draw(w);

	apply_iv_image(d, TRUE);

	d->freeze_count--;
}


/*
 *	Histogram GtkDrawingArea event signal callback.
 */
static gint histogram_event_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gint status = FALSE;
	gboolean key_press;
	guint keyval;
	GdkEventConfigure *configure;
	GdkEventFocus *focus;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	Dialog *d = DIALOG(data);
	if((widget == NULL) || (event == NULL) || (d == NULL))
	    return(status);

	if(d->freeze_count > 0)
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_CONFIGURE:
	    configure = (GdkEventConfigure *)event;
	    GDK_PIXMAP_UNREF(d->histogram_pm);
	    d->histogram_pm = gdk_pixmap_new(
		widget->window, configure->width, configure->height, -1
	    );
	    gtk_widget_queue_draw(widget);
	    status = TRUE;
	    break;

	  case GDK_EXPOSE:
	    histogram_draw(d);
	    status = TRUE;
	    break;

	  case GDK_FOCUS_CHANGE:
	    focus = (GdkEventFocus *)event;
	    if(focus->in && !GTK_WIDGET_HAS_FOCUS(widget))
	    {
		GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_FOCUS);
		gtk_widget_queue_draw(widget);
		status = TRUE;
	    }
	    else if(!focus->in && GTK_WIDGET_HAS_FOCUS(widget)) 
	    {
		GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);
		gtk_widget_queue_draw(widget);
		status = TRUE;
	    }
	    break;

	  case GDK_KEY_PRESS:
	  case GDK_KEY_RELEASE:
	    key = (GdkEventKey *)event;
	    key_press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	    keyval = key->keyval;
#define STOP_EMIT(_w_)				\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(_w_),				\
  key_press ?					\
   "key_press_event" : "key_release_event"	\
 );
	    switch(keyval)
	    {




	    }

#undef STOP_EMIT
	    break;

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    if(GTK_WIDGET_CAN_FOCUS(widget) && !GTK_WIDGET_HAS_FOCUS(widget))
		gtk_widget_grab_focus(widget);
	    d->drag_button = button->button;
	    d->drag_x = (gint)button->x;
	    d->drag_y = (gint)button->y;
	    /* Grab pointer */
	    if(!gdk_pointer_is_grabbed())
	    {
		const gint	x = (gint)button->x,
				x_start = (widget->allocation.width - 256) / 2;
		gint min, mid, max, d_min, d_mid, d_max;
		get_marker_values_in_bytes(d, &min, &mid, &max);

		/* Calculate the absolute distance from the pointer to
		 * each marker
		 */
		d_min = x - (x_start + min);
		if(d_min < 0)
		    d_min *= -1;
		d_mid = x - (x_start + mid);
		if(d_mid < 0)
		    d_mid *= -1;
		d_max = x - (x_start + max);
		if(d_max < 0)
		    d_max *= -1;

		/* See which marker we are closest to */
		if(d_max < d_min)
		    d->drag_marker = (d_max < d_mid) ? 2 : 1;
		else
		    d->drag_marker = (d_min < d_mid) ? 0 : 1;

		set_marker(d, d->drag_marker, x - x_start);
		gtk_widget_queue_draw(widget);
		queue_apply_iv_image(d);

		gdk_pointer_grab(
		    widget->window,
		    FALSE,
		    GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK,
		    NULL,
		    d->translate_cur,
		    button->time
		);
	    }
	    status = TRUE;
	    break;

	  case GDK_BUTTON_RELEASE:
	    button = (GdkEventButton *)event;
	    if(d->drag_button > 0)
	    {
		const gint x_start = (widget->allocation.width - 256) / 2;
		const gint x = (gint)button->x;

		set_marker(d, d->drag_marker, x - x_start);
		gtk_widget_queue_draw(widget);
		apply_iv_image(d, FALSE);	/* Apply non-interruptable */

		d->drag_button = 0;
		d->drag_x = 0;
		d->drag_y = 0;
		d->drag_marker = 0;
	    }
	    /* Ungrab pointer */
	    if(gdk_pointer_is_grabbed())
		gdk_pointer_ungrab(button->time);
	    status = TRUE;
	    break;

	  case GDK_MOTION_NOTIFY:
	    motion = (GdkEventMotion *)event;
	    if(d->drag_button == 1)
	    {
		const gint x_start = (widget->allocation.width - 256) / 2;
		gint x, y;
		GdkModifierType mask;
		GdkWindow *window = motion->window;

		if(motion->is_hint)
		{
		    gdk_window_get_pointer(
			window, &x, &y, &mask
		    );
		}
		else
		{
		    x = (gint)motion->x;
		    y = (gint)motion->y;
		}
		d->drag_x = x;
		d->drag_y = y;

		set_marker(d, d->drag_marker, x - x_start);
		gtk_widget_queue_draw(widget);
		queue_apply_iv_image(d);		
	    }
	    status = TRUE;
	    break;
	}

	return(status);
}

/*
 *	Make histogram idle callback.
 */
static gint make_histogram_idlecb(gpointer data)
{
	imgview_image_struct *img;
	Dialog *d = DIALOG(data);
	if(d == NULL)
	    return(FALSE);

	img = ImgViewGetImage(d->iv);
	if((img == NULL) || (d->orig_img_data == NULL))
	    return(FALSE);

	/* Create histogram */
	make_histogram(
	    d->histogram_data,
	    &d->histogram_data_max,
	    d->orig_img_data,
	    img->width, img->height, img->bpp, img->bpl
	);

	gtk_widget_queue_draw(d->histogram_da);

	return(FALSE);
}

/*
 *	Apply to ImgView timeout callback.
 */
static gint apply_iv_image_tocb(gpointer data)
{
	Dialog *d = DIALOG(data);
	if(d == NULL)
	    return(FALSE);

	d->apply_to_iv_toid = 0;

	/* Apply the color levels to the ImgView's current frame and
	 * allow interruption
	 */
	apply_iv_image(d, TRUE);

	return(FALSE);
}


/*
 *	Gets the marker values in units of bytes.
 */
static void get_marker_values_in_bytes(
	Dialog *d, gint *min_rtn, gint *mid_rtn, gint *max_rtn
)
{
	const gint min = gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(d->min_spin)
	),
	max = MAX(gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(d->max_spin)
	), min),
	mid = (gint)(
	    (gfloat)min + gtk_spin_button_get_value_as_float(
		GTK_SPIN_BUTTON(d->mid_spin)
	    ) / 2.0f * (gfloat)(max - min)
	);

	if(min_rtn != NULL)
	    *min_rtn = min;
	if(mid_rtn != NULL)
	    *mid_rtn = mid;
	if(max_rtn != NULL)
	    *max_rtn = max;
}

/*
 *	Sets the marker to the value specified by i but does not
 *	apply the changes to the current ImgView image frame.
 */
static void set_marker(
	Dialog *d, gint marker, gint i
)
{
	d->freeze_count++;

	/* Mid marker? */
	if(marker == 1)
	{
	    gint min, mid, max;
	    gfloat mid_coeff;

	    get_marker_values_in_bytes(d, &min, &mid, &max);

	    mid_coeff = (gfloat)(i - min) / (gfloat)MAX((max - min), 1);

	    gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(d->mid_spin),
		mid_coeff * 2.0f
	    );
	}
	else
	{
	    /* Max marker? */
	    if(marker == 2)
		gtk_spin_button_set_value(
		    GTK_SPIN_BUTTON(d->max_spin),
		    (gfloat)i
		);
	    else
		gtk_spin_button_set_value(
		    GTK_SPIN_BUTTON(d->min_spin),
		    (gfloat)i
		);
	}

	d->freeze_count--;
}



/*
 *	Redraws the Histogram GtkDrawingArea.
 */
static void histogram_draw(Dialog *d)
{
	gint i, x, y, width, height, min, mid, max;
	gfloat coeff, m;
	GdkPixmap *pixmap;
	GdkWindow *window;
	GdkDrawable *drawable;
	GdkGC *gc;
	GtkStateType state;
	GtkStyle *style;
	GtkWidget *w;

	if(d == NULL)
	    return;

	gc = d->gc;
	w = d->histogram_da;
	if((gc == NULL) || (w == NULL))
	    return;

	window = w->window;
	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
	if((window == NULL) || (style == NULL))
	    return;

	pixmap = d->histogram_pm;
	drawable = (GdkDrawable *)((pixmap != NULL) ? pixmap : window);
	gdk_window_get_size(drawable, &width, &height);
	if((width <= 0) || (height <= 0))
	    return;

	/* Draw background */
	gdk_draw_rectangle(
	    drawable, style->base_gc[state], TRUE,
	    0, 0, width, height
	);

	/* Get bounding positions in units of bytes */
	get_marker_values_in_bytes(d, &min, &mid, &max);

	/* Calculate x starting position so that the histogram is
	 * centered
	 */
	x = (width - 256) / 2;

	/* Get highest histogram value */
	m = MAX(d->histogram_data_max, 1.0f);

	/* Draw histogram */
	for(i = 0; i < 256; i++)
	{
	    coeff = (gfloat)d->histogram_data[i] / (gfloat)m;

	    y = height - (gint)(height * coeff);

	    /* Draw value line */
	    gdk_draw_line(
		drawable, style->fg_gc[state],
		x, y, x, height
	    );

	    x++;
	}

	/* Draw the min, mid, and max markers */
	x = (width - 256) / 2;
	gdk_gc_set_function(gc, GDK_INVERT);
	gdk_draw_line(
	    drawable, gc,
	    x + min, 0,
	    x + min, height
	);
	gdk_draw_line(
	    drawable, gc,
	    x + mid, 0,  
	    x + mid, height
	);
	gdk_draw_line(
	    drawable, gc,
	    x + max, 0,  
	    x + max, height
	);

	/* Send drawable to window if drawable is not the window */
	if(drawable != window)
	    gdk_draw_pixmap(
		window, style->fg_gc[state], drawable,
		0, 0, 0, 0, width, height
	    );
}


/*
 *	Applies color leveling to the specified pixel.
 */
static guint8 color_level_image_iterate(
	const gint v,
	const gint min_displace, const gint mid, const gint max_displace,
	const gint min_to_mid, const gint mid_to_max,
	const gint max_minus_mid
)
{
#if 0
	if(v < 0x80)
	{
	    const gint v2 = v * mid / 0x80;
	    const gint shift = min_displace * (mid - v2) / min_to_mid;
	    return((guint8)MAX(v2 - shift, 0));
	}
	else
	{
	    const gint v2 = ((v - 0x80) * max_minus_mid / 0x80) + mid;
	    const gint shift = max_displace * (v2 - mid) / mid_to_max;
	    return((guint8)MIN(v2 + shift, (gint)((guint8)-1)));
	}
#elif 0
	if(v < mid)
	{
	    const gint shift = min_displace * (mid - v) / min_to_mid;
	    return(MAX((v - shift), 0) * mid / 0x80);
	}
	else
	{
	    const gint shift = max_displace * (v - mid) / mid_to_max;
	    const gint v2 = MIN((v + shift), (gint)((guint8)-1));
	    return(((v2 - 0x80) * max_minus_mid / 0x80) + mid);
	}
#else
	/* Shift the value depending on if it is below or above the
	 * midpoint
	 */
	if(v < mid)
	{
	    const gint shift = min_displace * (mid - v) / min_to_mid;
	    return((guint8)MAX(v - shift, 0));
	}
	else
	{
	    const gint shift = max_displace * (v - mid) / mid_to_max;
	    return((guint8)MIN(v + shift, (gint)((guint8)-1)));
	}
#endif
}

/*
 *	Applies color leveling from the source image data to the
 *	target image data.
 */
static void color_level_image(
	const guint8 *src, guint8 *tar,
	const gint width, const gint height,    
	const gint bpp, const gint bpl,     
	gint min, gint mid, gint max,
	gboolean interruptable
)
{
	gint		y,
			min_to_mid, mid_to_max,
			min_displace, max_displace,
			max_minus_mid;
	const guint8 *src_ptr, *src_end;
	guint8 *tar_ptr;

	if((src == NULL) || (tar == NULL))
	    return;

	if(mid < min)
	    mid = min;
	if(max < mid)
	    max = mid;

	min_to_mid = mid - min;
	mid_to_max = max - mid;

	/* Absolute distances to mid must be positive or else we get
	 * an divid by 0 error
	 */
	if((min_to_mid <= 0) || (mid_to_max <= 0))
	    return;

	max_minus_mid = max - mid;

	min_displace = min;
	max_displace = (gint)((guint8)-1) - max;

	switch(bpp)
	{
	  case 4:	/* RGBA */
	    for(y = 0; y < height; y++)
	    {
		src_ptr = src + (y * bpl);
		src_end = src_ptr + (width * bpp);
		tar_ptr = tar + (y * bpl);
		while(src_ptr < src_end)
		{
		    *tar_ptr++ = color_level_image_iterate(
			(const gint)*src_ptr++,
			min_displace, mid, max_displace,
			min_to_mid, mid_to_max,
			max_minus_mid
		    );
		    *tar_ptr++ = color_level_image_iterate(
			(const gint)*src_ptr++,
			min_displace, mid, max_displace,
			min_to_mid, mid_to_max,
			max_minus_mid
		    );
		    *tar_ptr++ = color_level_image_iterate(
			(const gint)*src_ptr++,
			min_displace, mid, max_displace,
			min_to_mid, mid_to_max,
			max_minus_mid
		    );
		    *tar_ptr++ = *src_ptr++;
		}
		if(interruptable && (gtk_events_pending() > 0))
		    break;
	    }
	    break;

	  case 3:	/* RGB */
	    for(y = 0; y < height; y++)
	    {
		src_ptr = src + (y * bpl);
		src_end = src_ptr + (width * bpp);
		tar_ptr = tar + (y * bpl);
		while(src_ptr < src_end)
		{
		    *tar_ptr++ = color_level_image_iterate(
			(const gint)*src_ptr++,
			min_displace, mid, max_displace,
			min_to_mid, mid_to_max,
			max_minus_mid
		    );
		    *tar_ptr++ = color_level_image_iterate(
			(const gint)*src_ptr++,
			min_displace, mid, max_displace,
			min_to_mid, mid_to_max,
			max_minus_mid
		    );
		    *tar_ptr++ = color_level_image_iterate(
			(const gint)*src_ptr++,
			min_displace, mid, max_displace,
			min_to_mid, mid_to_max,
			max_minus_mid
		    );
		}
		if(interruptable && (gtk_events_pending() > 0))
		    break;
	    }
	    break;

	  case 2:	/* GA */
	    for(y = 0; y < height; y++)
	    {
		src_ptr = src + (y * bpl);
		src_end = src_ptr + (width * bpp);
		tar_ptr = tar + (y * bpl);
		while(src_ptr < src_end)
		{
		    *tar_ptr++ = color_level_image_iterate(
			(const gint)*src_ptr++,
			min_displace, mid, max_displace,
			min_to_mid, mid_to_max,
			max_minus_mid
		    );
		    *tar_ptr++ = *src_ptr++;
		}
		if(interruptable && (gtk_events_pending() > 0))
		    break;
	    }
	    break;

	  case 1:	/* G */
	    for(y = 0; y < height; y++)
	    {
		src_ptr = src + (y * bpl);
		src_end = src_ptr + (width * bpp);
		tar_ptr = tar + (y * bpl);
		while(src_ptr < src_end)
		{
		    *tar_ptr++ = color_level_image_iterate(
			(const gint)*src_ptr++,
			min_displace, mid, max_displace,
			min_to_mid, mid_to_max,
			max_minus_mid
		    );
		}
		if(interruptable && (gtk_events_pending() > 0))
		    break;
	    }
	    break;
	}
}

/*
 *	Applies the color levels to the ImgView's current frame
 *	and redraws the ImgView.
 *
 *	If interruptable is TRUE then the call to color_level_image()
 *	function will return before completely applying the color levels
 *	to the ImgView's current frame if/when pending GTK events are
 *	detected.
 */
static void apply_iv_image(Dialog *d, gboolean interruptable)
{
	gint min, mid, max;
	imgview_frame_struct *frame;
	imgview_image_struct *img;
	imgview_struct *iv;

	if(d == NULL)
	    return;

	iv = d->iv;
	img = ImgViewGetImage(iv);
	frame = ImgViewImageGetFrame(img, d->cur_frame);
	if(frame == NULL)
	    return;

	/* Get the current marker values */
	get_marker_values_in_bytes(d, &min, &mid, &max);

	/* Apply the color levels to the current ImgView's frame */
	color_level_image(
	    d->orig_img_data, frame->buf,
	    img->width, img->height,
	    img->bpp, img->bpl,
	    min, mid, max,
	    interruptable
	);

	/* Redraw the ImgView */
	ImgViewQueueDrawView(iv);
}

/*
 *	Queues a call to apply_iv_image() that is interruptable.
 */
static void queue_apply_iv_image(Dialog *d)
{
	if(d == NULL)
	    return;

	GTK_TIMEOUT_REMOVE(d->apply_to_iv_toid);
	d->apply_to_iv_toid = gtk_timeout_add(
	    400l,
	    apply_iv_image_tocb, d
	);
}

/*
 *	Restores the ImgView's current frame and redraws the ImgView.
 */
static void restore_iv_image(Dialog *d)
{
	imgview_frame_struct *frame;
	imgview_image_struct *img;
	imgview_struct *iv;

	if(d == NULL)
	    return;

	iv = d->iv;
	img = ImgViewGetImage(iv);
	frame = ImgViewImageGetFrame(img, d->cur_frame);
	if(frame == NULL)                               
	    return;

	/* Copy our original image data to the ImgView's current
	 * frame
	 */
	if((d->orig_img_data != NULL) && (frame->buf != NULL))
	    memcpy(
		frame->buf, d->orig_img_data,
		img->bpl * img->height
	    );

	/* Redraw the ImgView */
	ImgViewQueueDrawView(iv);
}


/*
 *	Creates a histogram from the specified image data.
 */
static void make_histogram(
	guint *histogram_data,		/* 256 guints */
	guint *histogram_data_max,
	const guint8 *buf,
	const gint width, const gint height,
	const gint bpp, const gint bpl
)
{
	gint y;
	guint8 g;
	const guint8 *src_ptr, *src_end;

	if((histogram_data == NULL) || (histogram_data_max == NULL) ||
	   (buf == NULL)
	)
	    return;

	/* Clear histogram */
	memset(histogram_data, 0x00, 256 * sizeof(guint));

/* Adds the histogram value at the specified index _i_ and updates
 * *histogram_data_max as needed
 */
#define INC_HISTOGRAM(_i_)	{			\
 const guint v = histogram_data[(_i_)] + 1;	\
 if(v > *histogram_data_max)			\
  *histogram_data_max = v;			\
 histogram_data[(_i_)] = v;			\
}

	/* Get histogram values */
	switch(bpp)
	{
	  case 4:	/* RGBA */
	    for(y = 0; y < height; y++)
	    {
		src_ptr = buf + (y * bpl);
		src_end = src_ptr + (width * bpp);
		while(src_ptr < src_end)
		{
		    if(src_ptr[3] < 0x80)
		    {
			src_ptr += bpp;
			continue;
		    }

		    g = *src_ptr++;	/* Red */
		    INC_HISTOGRAM(g);

		    g = *src_ptr++;	/* Green */
		    INC_HISTOGRAM(g);

		    g = *src_ptr++;	/* Blue */
		    INC_HISTOGRAM(g);

		    src_ptr++;		/* Alpha */
		}
	    }
	    break;

	  case 3:	/* RGB */
	    for(y = 0; y < height; y++)
	    {   
		src_ptr = buf + (y * bpl);
		src_end = src_ptr + (width * bpp);
		while(src_ptr < src_end)
		{
		    g = *src_ptr++;	/* Red */
		    INC_HISTOGRAM(g);

		    g = *src_ptr++;	/* Green */
		    INC_HISTOGRAM(g);

		    g = *src_ptr++;	/* Blue */
		    INC_HISTOGRAM(g);
		}
	    }
	    break;

	  case 2:	/* GA */
	    for(y = 0; y < height; y++)
	    {
		src_ptr = buf + (y * bpl);
		src_end = src_ptr + (width * bpp);
		while(src_ptr < src_end)
		{
		    if(src_ptr[1] < 0x80)
		    {
			src_ptr += bpp;
			continue;
		    }

		    g = *src_ptr++;	/* Grey */
		    INC_HISTOGRAM(g);

		    src_ptr++;		/* Alpha */
		}
	    }
	    break;

	  case 1:	/* G */
	    for(y = 0; y < height; y++)
	    {
		src_ptr = buf + (y * bpl);
		src_end = src_ptr + (width * bpp);
		while(src_ptr < src_end)
		{
		    g = *src_ptr++;	/* Grey */
		    INC_HISTOGRAM(g);
		}
	    }
	    break;
	}
#undef INC_HISTOGRAM
}


/*
 *	Builds a standard toplevel dialog and returns the pointers to
 *	the created widgets.
 */
static void build_std_dialog(
	gint width, gint height, const gchar *title,
	GtkWidget *ref_toplevel,
	GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
	GtkAccelGroup **accelgrp_rtn
)
{
	gint	bw = GUI_BUTTON_HLABEL_WIDTH_DEF,
		bh = GUI_BUTTON_HLABEL_HEIGHT_DEF,
		border_major = 5;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent, *parent2;
	GtkWidget *toplevel, *main_vbox;


	/* Create keyboard accelerator group */
	accelgrp = gtk_accel_group_new();
	if(accelgrp_rtn != NULL)
	    *accelgrp_rtn = accelgrp;

	/* Create toplevel */
	toplevel = parent = w = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_set_usize(w, width, height);
	gtk_window_set_title(GTK_WINDOW(w), title);
#ifdef PROG_NAME
	gtk_window_set_wmclass(
	    GTK_WINDOW(w), "dialog", PROG_NAME
	);
#endif
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_color_levels_48x48_xpm);
	}
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(delete_event_cb), toplevel
	);
	gtk_container_border_width(GTK_CONTAINER(w), 0);
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	if((ref_toplevel != NULL) ?
	    GTK_IS_WINDOW(GTK_OBJECT(ref_toplevel)) : FALSE
	)
	{
	    gtk_window_set_modal(GTK_WINDOW(toplevel), TRUE);
	    gtk_window_set_transient_for(
		GTK_WINDOW(toplevel), GTK_WINDOW(ref_toplevel)
	    );
	}
	if(toplevel_rtn != NULL)
	    *toplevel_rtn = toplevel;

	main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;
	if(client_vbox_rtn != NULL)
	    *client_vbox_rtn = w;

	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Buttons */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Set Button */
	w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_ok_20x20_xpm,
#if defined(PROG_LANGUAGE_SPANISH)
            "Conjunto",
#elif defined(PROG_LANGUAGE_FRENCH)
            "Serie",
#elif defined(PROG_LANGUAGE_GERMAN)
            "Satz",
#elif defined(PROG_LANGUAGE_ITALIAN)
            "Serie",
#elif defined(PROG_LANGUAGE_DUTCH)
            "Het Stel",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
            "Jogo",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
            "Sett",
#else
            "Set",
#endif
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(set_cb), toplevel
	);
	gtk_accel_group_add(
	    accelgrp, GDK_s, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_s);
	gtk_widget_show(w);

	/* Cancel Button */
	w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_cancel_20x20_xpm,
#if defined(PROG_LANGUAGE_SPANISH)
            "Cancele",
#elif defined(PROG_LANGUAGE_FRENCH)
            "Annuler",
#elif defined(PROG_LANGUAGE_GERMAN)
            "Heben",
#elif defined(PROG_LANGUAGE_ITALIAN)
            "Annullar",
#elif defined(PROG_LANGUAGE_DUTCH)
            "Annuler",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
            "Cancela",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
            "Kanseler",
#else
            "Cancel",
#endif
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(cancel_cb), toplevel
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_c, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_c);
	gtk_widget_show(w);
}


/*
 *	Queries user to edit color levels.
 *
 *	Returns TRUE on Set or FALSE on Cancel.
 */
gboolean iv_color_levels(
	GtkWidget *ref_toplevel,
	imgview_struct *iv
)
{
	const gint	border_major = 5,
			border_minor = 2;
	GdkWindow *window;
	GtkAccelGroup *accelgrp = NULL;
	GtkWidget *toplevel = NULL, *client_vbox = NULL;
	imgview_image_struct *img;
	imgview_frame_struct *frame;
	Dialog *d;

	/* Reset globals */
	got_response = FALSE;

	if(iv == NULL)
	    return(FALSE);

	/* Create the dialog */
	d = DIALOG(g_malloc0(sizeof(Dialog)));
	build_std_dialog(
	    400, -1,
	    "Color Levels",
	    ref_toplevel,
	    &toplevel, &client_vbox, &accelgrp
	);
	window = (toplevel != NULL) ? toplevel->window : NULL;
	d->toplevel = toplevel;
	d->accelgrp = accelgrp;
	d->gc = gdk_gc_new(window);
	d->freeze_count = 0;
	d->iv = iv;
	d->histogram_pm = NULL;
	memset(d->histogram_data, 0x00, 256 * sizeof(guint));
	d->histogram_data_max = 0;
	d->orig_img_data = NULL;
	d->cur_frame = ImgViewGetCurrentFrame(iv);
	d->drag_button = 0;
	d->drag_x = 0;
	d->drag_y = 0;
	d->drag_marker = 0;
	d->apply_to_iv_toid = 0;

	d->freeze_count++;

	/* Create a copy of the current frame */
	img = ImgViewGetImage(iv);
	frame = ImgViewImageGetFrame(img, d->cur_frame);
	if(frame != NULL)
	{
	    d->orig_img_data = (guint8 *)g_malloc(
		img->bpl * img->height
	    );
	    if((d->orig_img_data != NULL) && (frame->buf != NULL))
		memcpy(
		    d->orig_img_data, frame->buf,
		    img->bpl * img->height
		);
	}

	/* Create the dialog's widgets */
	if(client_vbox != NULL)
	{
	    gint value_width;
	    GdkWindow *window = toplevel->window;
	    GtkAdjustment *adj;
	    GtkStyle *style = gtk_widget_get_style(toplevel);
	    GtkWidget	*w,
			*parent = client_vbox,
			*parent2, *parent3, *parent4;

	    d->translate_cur = gdk_cursor_new(GDK_SB_H_DOUBLE_ARROW);

	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent = w;


	    /* Left column vbox */
	    w = gtk_vbox_new(TRUE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;
	    /* Icon */
	    w = gtk_pixmap_new_from_xpm_d(
		window,
		(style != NULL) ? &style->bg[GTK_STATE_NORMAL] : NULL,
		(guint8 **)icon_color_levels_32x32_xpm
	    );
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);


	    /* Right column vbox */
	    w = gtk_vbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	    gtk_widget_show(w);
	    parent2 = w;

	    w = gtk_frame_new(NULL);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Histogram GtkDrawingArea */
	    d->histogram_da = w = gtk_drawing_area_new();
	    GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS);
	    gtk_widget_set_usize(w, -1, 100);
	    gtk_widget_add_events(
		w,
		GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
		GDK_FOCUS_CHANGE_MASK |
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "configure_event",
		GTK_SIGNAL_FUNC(histogram_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "expose_event",
		GTK_SIGNAL_FUNC(histogram_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "focus_in_event",
		GTK_SIGNAL_FUNC(histogram_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "focus_out_event",
		GTK_SIGNAL_FUNC(histogram_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(histogram_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(histogram_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(histogram_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(histogram_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "motion_notify_event",
		GTK_SIGNAL_FUNC(histogram_event_cb), d
	    );
	    gtk_container_add(GTK_CONTAINER(parent3), w);
	    gtk_widget_realize(w);
	    gtk_widget_show(w);


	    /* GtkHBox for the min, mid, and max GtkSpinButtons */
	    w = gtk_hbox_new(TRUE, 0);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    value_width = 60;

	    /* Min */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    w = gtk_label_new("Min:");
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w); 
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, -256.0f, 255.0f, 1.0f, 10.0f, 0.0f  
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(adj_value_changed_cb), d
	    );
	    d->min_spin = w = gtk_spin_button_new(adj, 1.0, 0);
	    gtk_widget_set_usize(w, value_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    /* Mid */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    w = gtk_label_new("Mid:");
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		1.0f, 0.0f, 2.0f, 0.01f, 0.1f, 0.0f
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(adj_value_changed_cb), d
	    );
	    d->mid_spin = w = gtk_spin_button_new(adj, 1.0, 2);
	    gtk_widget_set_usize(w, value_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    /* Max */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    w = gtk_label_new("Max:");
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		255.0f, 0.0f, 511.0f, 1.0f, 10.0f, 0.0f
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(adj_value_changed_cb), d
	    );
	    d->max_spin = w = gtk_spin_button_new(adj, 1.0, 0);
	    gtk_widget_set_usize(w, value_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	}

	d->freeze_count--;

	/* Map toplevel */
	gtk_widget_show_raise(toplevel);

	/* Schedual idle call to make histogram */
	gtk_idle_add(make_histogram_idlecb, d);

	/* Push main loop to wait for user response */
	gtk_main();


	/* Get values from widgets if user clicked on save */
	if(got_response)
	{
	    apply_iv_image(d, FALSE);
	}
	else
	{
	    restore_iv_image(d);
	}

	/* Destroy widgets and related resources */
	GTK_TIMEOUT_REMOVE(d->apply_to_iv_toid);
	d->apply_to_iv_toid = 0;

	gtk_window_set_modal(GTK_WINDOW(toplevel), FALSE);
	gtk_window_set_transient_for(GTK_WINDOW(toplevel), NULL);

	GTK_WIDGET_DESTROY(d->min_spin);
	GTK_WIDGET_DESTROY(d->mid_spin);
	GTK_WIDGET_DESTROY(d->max_spin);
	GTK_WIDGET_DESTROY(d->histogram_da);
	GTK_WIDGET_DESTROY(d->toplevel);
	GTK_ACCEL_GROUP_UNREF(d->accelgrp);
	GDK_GC_UNREF(d->gc);
	GDK_PIXMAP_UNREF(d->histogram_pm);
	GDK_CURSOR_DESTROY(d->translate_cur);
	g_free(d->orig_img_data);
	g_free(d);

	/* Return user response */
	return(got_response);
}
