#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "imgio.h"
#include "guiutils.h"
#include "guirgbimg.h"
#include "cdialog.h"
#include "fb.h"
#include "pdialog.h"

#include "imgview.h"
#include "framesdlg.h"
#include "config.h"

#include "images/icon_animate_48x48.xpm"
#include "images/icon_animate_32x32.xpm"

#include "images/icon_add_20x20.xpm"
#include "images/icon_edit_20x20.xpm"
#include "images/icon_remove_20x20.xpm"
#include "images/icon_play_20x20.xpm"
#include "images/icon_pause_20x20.xpm"
#include "images/icon_close_20x20.xpm"


typedef struct _Dialog			Dialog;
typedef struct _FrameData		FrameData;


/*
 *	Image fitting types:
 */
typedef enum {
	IMG_FIT_RESIZE,			/* Resize and do not maintain aspect */
	IMG_FIT_RESIZE_MAINTAIN_ASPECT,
	IMG_FIT_RESIZE_MAINTAIN_ASPECT_TO_COVER,
	IMG_FIT_CENTERED,
	IMG_FIT_UPPER_LEFT,
	IMG_FIT_UPPER_RIGHT,
	IMG_FIT_LOWER_RIGHT,
	IMG_FIT_LOWER_LEFT
} img_fit_type;
#define IMG_FIT_STRINGS	{				\
	"Resize",					\
	"Resize & Maintain Aspect",			\
	"Resize & Maintain Aspect To Cover",		\
	"Centered",					\
	"Upper Left",					\
	"Upper Right",					\
	"Lower Right",					\
	"Lower Left",					\
	NULL						\
}

/*
 *	Dialog:
 */
struct _Dialog {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gint		freeze_count;
	imgview_struct	*iv;		/* Pointer back to the ImgView */

	GtkWidget	*clist,
			*add_btn,
			*edit_btn,
			*remove_btn,
			*shift_up_btn,
			*shift_down_btn,
			*play_btn,
			*pause_btn,
			*menu,
			*add_mi,
			*edit_mi,
			*remove_mi,
			*shift_up_mi,
			*shift_down_mi,
			*play_mi,
			*pause_mi;

	img_fit_type	img_fit;
};
#define DIALOG(p)	((Dialog *)(p))

/*
 *	Frame Data:
 */
struct _FrameData {

	gulong delay;

};
#define FRAME_DATA(p)	((FrameData *)(p))


/* Callbacks */
static gint delete_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void add_cb(GtkWidget *widget, gpointer data);
static void edit_cb(GtkWidget *widget, gpointer data);
static void remove_cb(GtkWidget *widget, gpointer data);
static void shift_up_cb(GtkWidget *widget, gpointer data);
static void shift_down_cb(GtkWidget *widget, gpointer data);
static void play_cb(GtkWidget *widget, gpointer data);
static void pause_cb(GtkWidget *widget, gpointer data);
static void close_cb(GtkWidget *widget, gpointer data);
static gint clist_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void clist_select_row_cb(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);
static void clist_unselect_row_cb(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);
static void clist_drag_data_received_cb(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
);
static gchar *pdialog_browse_cb(
	gpointer pdialog, gpointer data, gint prompt_num
);
static gint imgio_progress_cb(
	gpointer data,
	gint pi, gint pm,
	gint width, gint height,
	gint bpl, gint bpp,
	guint8 *rgba
);

/* GtkCList */
static void clist_row_update(
	GtkCList *clist, gint row,
	guint8 *rgba, gint width, gint height, gint bpl,
	gulong delay
);
static gint clist_row_insert(
	GtkCList *clist, gint row,
	guint8 *rgba, gint width, gint height, gint bpl,
	gulong delay
);

/* Frame image setting */
static void set_frame_from_image_file(
	Dialog *d,
	imgview_image_struct *img,
	imgview_frame_struct *frame,
	const gchar *path,
	img_fit_type img_fit
);

/* Frame ops */
static void add_frame(Dialog *d);
static void edit_frame(Dialog *d);
static void remove_frame(Dialog *d);
static void shift_frame_up(Dialog *d);
static void shift_frame_down(Dialog *d);

/* Dialog Updating */
static void update(Dialog *d);

/* Utils */
static FrameData *frame_data_new(
	gulong delay
);
static void frame_data_delete(FrameData *fd);
static img_fit_type img_fit_type_get_from_string(const gchar *s);
static gchar *img_fit_string(img_fit_type img_fit);

static void build_std_dialog(
	gint width, gint height, const gchar *title,
	GtkWidget *ref_toplevel,
	GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
	GtkAccelGroup **accelgrp_rtn
);

gboolean iv_edit_frames(
	GtkWidget *ref_toplevel,
	imgview_struct *iv,
	const gchar *filename
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)

/* Size of frame icon in the GtkCList */
#define FRAME_ICON_WIDTH		48
#define FRAME_ICON_HEIGHT		48

static gboolean		has_changes;


/*
 *	Toplevel GtkWindow "delete_event" signal callback.
 */
static gint delete_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gtk_main_quit();
	return(TRUE);
}

/*
 *	Add callback.
 */
static void add_cb(GtkWidget *widget, gpointer data)
{
	add_frame(DIALOG(data));
}

/*
 *	Edit callback.
 */
static void edit_cb(GtkWidget *widget, gpointer data)
{
	edit_frame(DIALOG(data));
}

/*
 *	Remove callback.
 */
static void remove_cb(GtkWidget *widget, gpointer data)
{
	remove_frame(DIALOG(data));
}

/*
 *	Shift up callback.
 */
static void shift_up_cb(GtkWidget *widget, gpointer data)
{
	shift_frame_up(DIALOG(data));
}

/*
 *	Shift down callback.
 */
static void shift_down_cb(GtkWidget *widget, gpointer data)
{
	shift_frame_down(DIALOG(data));
}

/*
 *	Play callback.
 */
static void play_cb(GtkWidget *widget, gpointer data)
{
	imgview_struct *iv;
	Dialog *d = DIALOG(data);
	if(d == NULL)
	    return;

	iv = d->iv;
	if(iv == NULL)
	    return;

	if(!ImgViewIsPlaying(iv) && (ImgViewGetTotalFrames(iv) > 1))
	    ImgViewPlay(iv);

	update(d);
}

/*
 *	Pause callback.
 */
static void pause_cb(GtkWidget *widget, gpointer data)
{
	imgview_struct *iv;
	Dialog *d = DIALOG(data);
	if(d == NULL)
	    return;

	iv = d->iv;
	if(iv == NULL)
	    return;

	if(ImgViewIsPlaying(iv))
	    ImgViewPause(iv);

	update(d);
}

/*
 *	Close callback.
 */
static void close_cb(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

/*
 *	GtkCList event signal callback.
 */
static gint clist_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventButton *button;
	Dialog *d = DIALOG(data);
	if((widget == NULL) || (event == NULL) || (d == NULL))
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    if(button->button == GDK_BUTTON3)
	    {
		GtkWidget *w = d->menu;
		if(w != NULL)
		    gtk_menu_popup(
			GTK_MENU(w), NULL, NULL,
			NULL, NULL,
			button->button, button->time
		    );
		status = TRUE;
	    }
	    break;
	}

	return(status);
}

/*
 *	GtkCList "select_row" signal callback.
 */
static void clist_select_row_cb(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	imgview_struct *iv;
	Dialog *d = DIALOG(data);
	if((clist == NULL) || (event == NULL) || (d == NULL))
	    return;

	iv = d->iv;
	if(iv == NULL)
	    return;

	/* Scroll to the selected row if it is not fully visible */
	if(gtk_clist_row_is_visible(clist, row) !=
	    GTK_VISIBILITY_FULL
	)
	    gtk_clist_moveto(
		clist,
		row, -1,	/* Row, column */
		0.5f, 0.0f	/* Row, column */
	    );

	/* Seek the ImgView to the selected frame if it is not
	 * playing
	 */
	if(!ImgViewIsPlaying(iv))
	    ImgViewSeek(iv, row);

	update(d);

	/* Edit on double-click */
	if(event->type == GDK_2BUTTON_PRESS)
	    edit_frame(d);
}

/*
 *	GtkCList "unselect_row" signal callback.
 */
static void clist_unselect_row_cb(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	Dialog *d = DIALOG(data);
	if((clist == NULL) || (event == NULL) || (d == NULL))
	    return;

	update(d);
}

/*
 *	GtkCList "drag_data_received" signal callback.
 */
static void clist_drag_data_received_cb(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
)
{
	gint row, column;
	GtkCList *clist = GTK_CLIST(widget);
	imgview_struct *iv;
	imgview_image_struct *img;
	Dialog *d = DIALOG(data);
	if((clist == NULL) || (dc == NULL) || (selection_data == NULL) ||
	   (d == NULL) || CDialogIsQuery() || FileBrowserIsQuery() ||
	   PDialogIsQuery()
	)
	    return;

	iv = d->iv;
	if((selection_data->data == NULL) || (iv == NULL))
	    return;

	img = ImgViewGetImage(iv);
	if(img == NULL)
	    return;

	/* Calculate the row and column based on the drop coordinates */
	if(!gtk_clist_get_selection_info(
	    clist,
	    x,
	    y - ((clist->flags & GTK_CLIST_SHOW_TITLES) ?
		clist->column_title_area.height +
		clist->column_title_area.y : 0),
	    &row, &column
	))
	{
	    row = -1;
	    column = 0;
	}

	/* Handle by data type */
	if((info == DND_INFO_TEXT_PLAIN) ||
	   (info == DND_INFO_TEXT_URI_LIST) ||
	   (info == DND_INFO_STRING)
	)
	{
	    const gchar *img_fit_strings[] = IMG_FIT_STRINGS;
	    gint i, strc;
	    const gchar	*pfx = "file://",
			*url = selection_data->data,
			*path;
	    gchar *mesg, **strv, *img_fit_s;
	    GList *img_fit_list;
	    GtkWidget *toplevel = d->toplevel;

	    /* Parse path from the url */
	    if(g_strcasepfx(url, pfx))
		path = url + STRLEN(pfx);
	    else
		path = url;

	    /* Query user for image fit type */
	    PDialogDeleteAllPrompts();
	    gtk_window_set_transient_for(
		GTK_WINDOW(PDialogGetToplevel()),
		GTK_WINDOW(toplevel)
	    );
	    PDialogSetSize(350, -1);
	    img_fit_list = NULL;
	    for(i = 0; img_fit_strings[i] != NULL; i++)
		img_fit_list = g_list_append(
		    img_fit_list, STRDUP(img_fit_strings[i])
		);
	    img_fit_s = img_fit_string(d->img_fit);
	    PDialogAddPromptCombo(
		NULL, "Fit Image:", img_fit_s,
		img_fit_list,
		FALSE, FALSE
	    );
	    PDialogSetPromptTip(
		0,
"Select the method in which the image will be resized and\
 positioned in the frame"
	    );
	    g_list_foreach(img_fit_list, (GFunc)g_free, NULL);
	    g_list_free(img_fit_list);
	    g_free(img_fit_s);
	    PDialogAddPromptSpin(  
		NULL, "Delay:",
		(gfloat)250.0f,
		0.0f,
		(gfloat)((guint)-1),	/* Reasonable max delay, right? */
		1.0f, 5.0f,
		1.0f, 0
	    );
            PDialogSetPromptTip(
		1,
"Set the delay in milliseconds which specifies how long this\
 frame will be displayed before displaying the next frame"
            );
	    mesg = g_strdup_printf("File: %s", g_basename(path));
	    strv = PDialogGetResponse(
		"Insert New Frame",
		mesg, NULL,
		PDIALOG_ICON_EDIT,
		"Insert", "Cancel",
		PDIALOG_BTNFLAG_SUBMIT | PDIALOG_BTNFLAG_CANCEL,
		PDIALOG_BTNFLAG_SUBMIT,
		&strc
	    );
	    g_free(mesg);
	    if(strv != NULL)
	    {
		const gchar *img_fit_s = (strc > 0) ? strv[0] : NULL;
		const gulong delay = (strc > 1) ? ATOL(strv[1]) : 250l;
		gint frame_num;
		imgview_frame_struct *frame;

		/* Get the image fit type */
		d->img_fit = img_fit_type_get_from_string(img_fit_s);

	        /* Mark has changes */
	        has_changes = TRUE;

	        /* Insert or append a new frame */
	        frame_num = ImgViewImageInsertFrame(
		    img, row,
		    (guint8 *)g_malloc0(img->bpl * img->height),
		    delay
	        );

	        /* Get new frame */
	        frame = ImgViewImageGetFrame(img, frame_num);
	        if(frame != NULL)
	        {
		    gint new_row;

		    /* Need to update the alpha channel flags on the
		     * ImgView to note that the alpha channel is
		     * defined because creating a new frame usually
		     * has transparency
		     */
		    if(!(iv->alpha_flags & IMGVIEW_ALPHA_DEFINED))
		    {
			iv->alpha_flags |= IMGVIEW_ALPHA_DEFINED;
		    }

		    /* Open the image specified by the path to the
		     * new frame
		     */
		    set_frame_from_image_file(
		        d, img, frame, path, d->img_fit
		    );

		    /* Insert clist row */
		    new_row = clist_row_insert(
		        clist, row,
		        frame->buf, img->width, img->height, img->bpl,
		        frame->delay
		    );

	            /* Select the new frame */
	            gtk_clist_freeze(clist);
	            gtk_clist_unselect_all(clist);
	            gtk_clist_select_row(clist, new_row, 0);
	            gtk_clist_thaw(clist);
	        }

		if(!ImgViewIsPlaying(iv))
		    ImgViewQueueDrawView(iv);
	        update(d);
	    }

	    PDialogSetTransientFor(NULL);
	    PDialogDeleteAllPrompts();
	}
}

/*
 *	PDialog browse callback.
 */
static gchar *pdialog_browse_cb(
	gpointer pdialog, gpointer data, gint prompt_num
)
{
	gboolean response;
	gint npaths = 0, total_ftypes = 0;
	gchar **paths_list;
	fb_type_struct *type_rtn, **ftype = NULL;
	static gchar path[PATH_MAX + NAME_MAX];

	if(FileBrowserIsQuery())
	    return(NULL);

	/* Create the file types list */
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "All Images"
	);

	*path = '\0';

	/* Need to set transient for manually to avoid setting modal */
	gtk_window_set_transient_for(
	    GTK_WINDOW(FileBrowserGetToplevel()),
	    GTK_WINDOW(PDialogGetToplevel())
	);

	/* Prompt user for image */
	response = FileBrowserGetResponse(
#if defined(PROG_LANGUAGE_SPANISH)
            "Escoja La Imagen",
            "Escoja", "Cancele",
#elif defined(PROG_LANGUAGE_FRENCH)
            "Choisir L'Image",
            "Privilgi", "Annuler",
#elif defined(PROG_LANGUAGE_GERMAN)
            "Wahlen Sie Bildnis Aus",
            "Wahlen", "Heben",
#elif defined(PROG_LANGUAGE_ITALIAN)
            "Scegliere L'Immagine",
            "Sceglier", "Annullar",
#elif defined(PROG_LANGUAGE_DUTCH)
            "Selecteer Beeld",
            "Selecter", "Annuler",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
            "Selecione Imagem",
            "Selecion", "Cancela",   
#elif defined(PROG_LANGUAGE_NORWEGIAN)
            "Velg Ut Avbilde",
            "Velg", "Kanseler",
#else
            "Select Image",
            "Select", "Cancel",
#endif
	    NULL,
	    ftype, total_ftypes,
	    &paths_list, &npaths,
	    &type_rtn
	);

        FileBrowserSetTransientFor(NULL);

	if(response)
	{
	    strncpy(
		path,
		paths_list[0],
		sizeof(path)
	    );
	    path[sizeof(path) - 1] = '\0';
	}

	/* Delete the file types list */
	FileBrowserDeleteTypeList(ftype, total_ftypes);

	return(path);
}

/*
 *	ImgIO progress callback.
 */
static gint imgio_progress_cb(
	gpointer data,
	gint pi, gint pm,
	gint width, gint height,
	gint bpl, gint bpp,
	guint8 *rgba
)
{
	return(TRUE);
}


/*
 *	Updates the GtkCList row with the specified frame data.
 */
static void clist_row_update(
	GtkCList *clist, gint row,
	guint8 *rgba, gint width, gint height, gint bpl,
	gulong delay
)
{
	const gint	iwidth = FRAME_ICON_WIDTH,
			iheight = FRAME_ICON_HEIGHT,
			bpp = 4;
	gchar *s;
	GdkWindow *window;
	GdkBitmap *mask;
	GdkPixmap *pixmap;
	GtkStyle *style;
	FrameData *fd;

	if(clist == NULL)
	    return;

	if(bpl <= 0)
	    bpl = width * bpp;

	window = GTK_WIDGET(clist)->window;
	style = gtk_widget_get_style(GTK_WIDGET(clist));
	if(style == NULL)
	    return;

	gtk_clist_freeze(clist);

	if((width > iwidth) || (height > iheight))
	{
	    const gfloat aspect = (height > 0) ?
		((gfloat)width / (gfloat)height) : 1.0f;
	    gint width2, height2;
	    guint8 *rgba2;

	    width2 = iwidth;
	    height2 = (aspect > 0.0f) ? (gint)(iheight / aspect) : iheight;
	    if(height2 > iheight)
	    {
		height2 = iheight;
		width2 = (gint)(iwidth * aspect);
	    }

	    rgba2 = (guint8 *)g_malloc(width2 * height2 * bpp);
	    GUIImageBufferResize(
		bpp,
		rgba, width, height, bpl,
		rgba2, width2, height2, width2 * bpp
	    );
	    pixmap = gdk_pixmap_new(window, width2, height2, -1);
	    if((pixmap != NULL) && (rgba2 != NULL))
		gdk_draw_rgb_32_image(
		    (GdkDrawable *)pixmap,
		    style->black_gc,
		    0, 0, width2, height2,
		    GDK_RGB_DITHER_NONE,
		    (guchar *)rgba2,
		    width2 * bpp
		);
	    mask = GUICreateBitmapFromDataRGBA(
		width2, height2, width2 * bpp,
		rgba2, 0x80,
		window
	    );
	    g_free(rgba2);
	}
	else
	{
	    pixmap = gdk_pixmap_new(window, width, height, -1);
	    if((pixmap != NULL) && (rgba != NULL))
		gdk_draw_rgb_32_image(
		    (GdkDrawable *)pixmap,
		    style->black_gc,
		    0, 0, width, height,
		    GDK_RGB_DITHER_NONE,
		    (guchar *)rgba,
		    bpl
		);
	    mask = GUICreateBitmapFromDataRGBA(
		width, height, bpl,
		rgba, 0x80,
		window
	    );
	}
	gtk_clist_set_pixmap(
	    clist, row, 0, pixmap, mask
	);
	GDK_PIXMAP_UNREF(pixmap);
	GDK_BITMAP_UNREF(mask);

	s = g_strdup_printf("%ld ms", delay);
	gtk_clist_set_text(
	    clist, row, 1, s
	);
	g_free(s);

	/* Update the frame data */
	fd = FRAME_DATA(gtk_clist_get_row_data(clist, row));
	if(fd != NULL)
	{
	    fd->delay = delay;
	}

	gtk_clist_thaw(clist);
}

/*
 *	Inserts a row to the GtkCList with the specified frame data.
 */
static gint clist_row_insert(
	GtkCList *clist, gint row,
	guint8 *rgba, gint width, gint height, gint bpl,
	gulong delay
)
{
	gint i, columns, new_row;
	gchar **strv;
	FrameData *fd;

	if(clist == NULL)
	    return(-1);

	gtk_clist_freeze(clist);

	/* Create a new row */
	columns = clist->columns;
	strv = (gchar **)g_malloc(columns * sizeof(gchar *));
	for(i = 0; i < columns; i++)
	    strv[i] = "";
	if(row > -1)
	    new_row = gtk_clist_insert(clist, row, strv);
	else
	    new_row = gtk_clist_append(clist, strv);
	g_free(strv);

	if(new_row < 0)
	{
	    gtk_clist_thaw(clist);
	    return(-1);
	}

	/* Set the new row */
	clist_row_update(
	    clist, new_row,
	    rgba, width, height, bpl,
	    delay
	);

	/* Set the new row's data */
	fd = frame_data_new(delay);
	gtk_clist_set_row_data_full(
	    clist, new_row,
	    fd, (GtkDestroyNotify)frame_data_delete
	);

	gtk_clist_thaw(clist);

	return(new_row);
}


/*
 *	Opens the image specified by path, crops it as needed,
 *	and sets it on the specified frame
 */
static void set_frame_from_image_file(
	Dialog *d,
	imgview_image_struct *img,
	imgview_frame_struct *frame,
	const gchar *path,
	img_fit_type img_fit
)
{
	gint		i, width, height, bpp, bpl,
			nframes,
			x, y, base_width, base_height;
	gchar		*creator,
			*title,
			*author,
			*comments;
	guint8 **rgba_list, bg_color[4];
	gulong *delay_list;
	GtkWidget *toplevel;

	if((d == NULL) || (img == NULL) || (frame == NULL) ||
	   STRISEMPTY(path)
	)
	    return;

	toplevel = d->toplevel;

	/* Open the image */
	ImgLoadFileRGBA(
	    path,
	    &width, &height, &bpp, &bpl,
	    &rgba_list, &delay_list, &nframes,
	    bg_color,
	    &x, &y, &base_width, &base_height,
	    &creator, &title, &author, &comments,
	    0xff,
	    d,
	    imgio_progress_cb
	);
	if((rgba_list != NULL) && (img->bpp == bpp))
	{
	    const guint8 *rgba = (nframes > 0) ? rgba_list[0] : NULL;
	    if((rgba != NULL) && (frame->buf != NULL))
	    {
		const gint	src_width = width,
				src_height = height,
				src_bpl = bpl,
				tar_width = img->width,
				tar_height = img->height,
				tar_bpl = img->bpl;
		const gfloat	aspect = (src_height > 0) ?
		    ((gfloat)src_width / (gfloat)src_height) : 1.0f;
		gint tmp_width, tmp_height;
		const guint8 *src = rgba;
		guint8 *tar = frame->buf;

		/* Clear the target frame first */
		memset(tar, 0x00, tar_bpl * tar_height);

		/* Copy the source image data to the target image
		 * data based on the specified image fit method
		 */
		switch(img_fit)
		{
		  case IMG_FIT_RESIZE:
		    GUIImageBufferResize(
			4,
			src, src_width, src_height, src_bpl,
			tar, tar_width, tar_height, tar_bpl
		    );
		    break;

		  case IMG_FIT_RESIZE_MAINTAIN_ASPECT:
		    tmp_width = tar_width;
		    tmp_height = (aspect > 0.0f) ?
			(gint)(tmp_width / aspect) : tmp_width;
		    if(tmp_height > tar_height)
		    {
			tmp_height = tar_height;
			tmp_width = (gint)(tmp_height * aspect);
		    }
		    if((tmp_width > 0) && (tmp_height > 0))
		    {
			const gint tmp_bpl = tmp_width * 4;
			guint8 *tmp = (guint8 *)g_malloc(
			    tmp_bpl * tmp_height
			);
			GUIImageBufferResize(
			    4,
			    src, src_width, src_height, src_bpl,
			    tmp, tmp_width, tmp_height, tmp_bpl
			);
			GUIImageBufferCopyArea(
			    4,
			    tmp, tmp_width, tmp_height, tmp_bpl,
			    tar, tar_width, tar_height, tar_bpl,
			    (tar_width - tmp_width) / 2,
			    (tar_height - tmp_height) / 2
			);
			g_free(tmp);
		    }
		    break;

		  case IMG_FIT_RESIZE_MAINTAIN_ASPECT_TO_COVER:
		    tmp_width = tar_width;
		    tmp_height = (aspect > 0.0f) ?
			(gint)(tmp_width / aspect) : tmp_width;
		    if(tmp_height < tar_height)
		    {
			tmp_height = tar_height;
			tmp_width = (gint)(tmp_height * aspect);
		    }
		    if((tmp_width > 0) && (tmp_height > 0))
		    {
			const gint tmp_bpl = tmp_width * 4;
			guint8 *tmp = (guint8 *)g_malloc(
			    tmp_bpl * tmp_height 
			);
			GUIImageBufferResize(
			    4,
			    src, src_width, src_height, src_bpl,
			    tmp, tmp_width, tmp_height, tmp_bpl
			);
			GUIImageBufferCopyArea(
			    4,
			    tmp, tmp_width, tmp_height, tmp_bpl,
			    tar, tar_width, tar_height, tar_bpl,
			    (tar_width - tmp_width) / 2,
			    (tar_height - tmp_height) / 2
			);
			g_free(tmp);
		    }
		    break;

		  case IMG_FIT_CENTERED:
		    GUIImageBufferCopyArea(
			4,
			src, src_width, src_height, src_bpl,
			tar, tar_width, tar_height, tar_bpl,
			(tar_width - src_width) / 2,
			(tar_height - src_height) / 2
		    );
		    break;

		  case IMG_FIT_UPPER_LEFT:
		    GUIImageBufferCopyArea(
			4,
			src, src_width, src_height, src_bpl,
			tar, tar_width, tar_height, tar_bpl,
			0, 0
		    );
		    break;

		  case IMG_FIT_UPPER_RIGHT:
		    GUIImageBufferCopyArea(
			4,
			src, src_width, src_height, src_bpl,
			tar, tar_width, tar_height, tar_bpl,
			tar_width - src_width,
			0
		    );
		    break;

		  case IMG_FIT_LOWER_RIGHT:
		    GUIImageBufferCopyArea(
			4,
			src, src_width, src_height, src_bpl,
			tar, tar_width, tar_height, tar_bpl,
			tar_width - src_width,
			tar_height - src_height
		    );      
		    break;

		  case IMG_FIT_LOWER_LEFT:
		    GUIImageBufferCopyArea(
			4,
			src, src_width, src_height, src_bpl,
			tar, tar_width, tar_height, tar_bpl,
			0,
			tar_height - src_height
		    );      
		    break;
		}
	    }
	}
	else
	{
	    /* Unable to open the image */
	    gchar *buf = g_strdup_printf(
"Unable to open image:\n\
\n\
    %s\n\
\n\
%s",
		path, ImgLoadGetError()
	    );
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Open Image Failed",
		buf, NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
	    );
	    g_free(buf);
	    CDialogSetTransientFor(NULL);
	}

	/* Delete the loaded image */
	for(i = 0; i < nframes; i++)
	    g_free(rgba_list[i]);
	g_free(rgba_list);
	g_free(delay_list);
	g_free(creator);
	g_free(title);
	g_free(author);
	g_free(comments);
}


/*
 *	Add frame.
 */
static void add_frame(Dialog *d)
{
	gint frame_num, row;
	GList *glist;
	GtkWidget *toplevel;
	GtkCList *clist;
	imgview_struct *iv;
	imgview_frame_struct *frame;
	imgview_image_struct *img;

	if(d == NULL)
	    return;

	toplevel = d->toplevel;
	iv = d->iv;
	if(iv == NULL)
	    return;

	img = ImgViewGetImage(iv);
	if(img == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(   
		"Add Frame Failed",
"There is no image loaded.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}

	clist = GTK_CLIST(d->clist);
	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;

	/* Mark has changes */
	has_changes = TRUE;

	/* Insert or append a new frame */
	frame_num = ImgViewImageInsertFrame(
	    img, row,
	    (guint8 *)g_malloc0(img->bpl * img->height),
	    250l
	);
	frame = ImgViewImageGetFrame(img, frame_num);
	if(frame == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Add Frame Failed",
"A new frame could not be inserted into the image.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}

	/* Need to update the alpha channel flags on the ImgView to
	 * note that the alpha channel is defined because creating
	 * a new frame usually has transparency
	 */
	if(!(iv->alpha_flags & IMGVIEW_ALPHA_DEFINED))
	{
	    iv->alpha_flags |= IMGVIEW_ALPHA_DEFINED;
	}

	/* Insert clist row */
	row = clist_row_insert(
	    clist, row,
	    frame->buf, img->width, img->height, img->bpl,
	    frame->delay
	);

	/* Select the new frame */
	gtk_clist_freeze(clist);
	gtk_clist_unselect_all(clist);
	gtk_clist_select_row(clist, row, 0);
	gtk_clist_thaw(clist);

	if(!ImgViewIsPlaying(iv))
	    ImgViewQueueDrawView(iv);
	update(d);

	/* Edit the new frame */
	edit_frame(d);
}

/*
 *	Edit frame.
 */
static void edit_frame(Dialog *d)
{
	const gchar *img_fit_strings[] = IMG_FIT_STRINGS;
	gint i, frame_num, strc;
	gchar *img_fit_s, *title, **strv;
	GList *glist, *img_fit_list;
	GtkWidget *toplevel;
	GtkCList *clist;
	imgview_struct *iv;
	imgview_frame_struct *frame;
	imgview_image_struct *img;

	if((d == NULL) || PDialogIsQuery())
	    return;

	toplevel = d->toplevel;
	iv = d->iv;
	if(iv == NULL)
	    return;

	img = ImgViewGetImage(iv);
	if(img == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Edit Frame Failed",
"There is no image loaded.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}

	clist = GTK_CLIST(d->clist);
	glist = clist->selection_end;
	frame_num = (glist != NULL) ? (gint)glist->data : -1;
	if(frame_num < 0)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Edit Frame Failed",
"There is no frame selected.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}

	/* Get selected frame */
	frame = ImgViewImageGetFrame(img, frame_num);
	if(frame == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Edit Frame Failed",
"Unable to get the selected frame from the image.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    return;
	}

	/* Get image fit strings */
	img_fit_list = NULL;
	for(i = 0; img_fit_strings[i] != NULL; i++)
	    img_fit_list = g_list_append(
		img_fit_list, STRDUP(img_fit_strings[i])
	    );
	img_fit_s = img_fit_string(d->img_fit);

	/* Set up the PDialog and prompt the user */
	PDialogDeleteAllPrompts();
	gtk_window_set_transient_for(
	    GTK_WINDOW(PDialogGetToplevel()),
	    GTK_WINDOW(toplevel)
	);
	PDialogSetSize(350, -1);
	PDialogAddPromptWithBrowse(
	    NULL,
	    "File:",
	    NULL,
	    d,
	    pdialog_browse_cb
	);
        PDialogSetPromptTip(
	    0,
"Select the image file to be placed into this frame, or\
 leave this blank to keep the current frame's data"
        );
	PDialogAddPromptCombo(
	    NULL, "Fit Image:", img_fit_s,
	    img_fit_list,
	    FALSE, FALSE
	);
        PDialogSetPromptTip(
	    1,
"Select the method in which the image will be resized and\
 positioned in the frame"
        );
	g_list_foreach(img_fit_list, (GFunc)g_free, NULL);
	g_list_free(img_fit_list);
	g_free(img_fit_s);
	PDialogAddPromptSpin(
	    NULL, "Delay:",
	    (gfloat)frame->delay,
	    0.0f,
	    (gfloat)((guint)-1),	/* Reasonable max delay, right? */
	    1.0f, 5.0f,
	    1.0f, 0
	);
        PDialogSetPromptTip(
	    2,
"Set the delay in milliseconds which specifies how long this\
 frame will be displayed before displaying the next frame"
        );
	title = g_strdup_printf("Frame #%i", frame_num + 1);
	strv = PDialogGetResponse(
	    title, NULL, NULL,
	    PDIALOG_ICON_EDIT,
	    "Set", "Cancel",
	    PDIALOG_BTNFLAG_SUBMIT | PDIALOG_BTNFLAG_CANCEL,
	    PDIALOG_BTNFLAG_SUBMIT,
	    &strc
	);
	g_free(title);
	if(strv != NULL)
	{
	    /* Got response, get new frame values */
	    const gchar *path = (strc > 0) ? strv[0] : NULL;
	    const gchar *img_fit_s = (strc > 1) ? strv[1] : NULL;
	    const gulong delay = (strc > 2) ? ATOL(strv[2]) : 0l;

	    /* Should also reget frame just in case */
	    imgview_frame_struct *frame = ImgViewImageGetFrame(
		img, frame_num
	    );

	    /* Get image fit type */
	    d->img_fit = img_fit_type_get_from_string(img_fit_s);

	    /* Selected frame still exists? */
	    if(frame != NULL)
	    {
		/* Set new frame values */
		frame->delay = delay;

		/* Set new image? */
		if(!STRISEMPTY(path))
		    set_frame_from_image_file(
			d, img, frame, path, d->img_fit
		    );

		/* Update the displayed row */
		clist_row_update(
		    clist, frame_num,
		    frame->buf, img->width, img->height, img->bpl,
		    frame->delay
		);

		/* Mark has changes */
		has_changes = TRUE;

		if(!ImgViewIsPlaying(iv))
		    ImgViewQueueDrawView(iv);
		update(d);
	    }
	    else
	    {
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
		    "Edit Frame Failed",
"The selected frame no longer exists.",
		    NULL,
		    CDIALOG_ICON_WARNING,
		    CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
	    }
	}

	PDialogSetTransientFor(NULL);
	PDialogDeleteAllPrompts();
}

/*
 *	Remove frame.
 */
static void remove_frame(Dialog *d)
{
	gint i, response;
	gchar *s;
	GtkWidget *toplevel;
	GtkCList *clist;
	imgview_struct *iv;
	imgview_image_struct *img;

	if((d == NULL) || CDialogIsQuery())
	    return;

	toplevel = d->toplevel;
	clist = GTK_CLIST(d->clist);
	iv = d->iv;
	if(iv == NULL)
	    return;

	img = ImgViewGetImage(iv);
	if(img == NULL)
	{
	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"Remove Frame Failed",
"There are no frames selected.",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);    
	    return;
	}

	/* Confirm remove */
	i = g_list_length(clist->selection);
	if(i == 1)
	    s = g_strdup_printf(
		"Remove frame #%i?",
		(gint)(clist->selection->data) + 1
	    );
	else
	    s = g_strdup_printf(
		"Remove %i selected frames?",
		i
	    );
	CDialogSetTransientFor(toplevel);
	response = CDialogGetResponse(
	    "Confirm Remove",
	    s, NULL,
	    CDIALOG_ICON_QUESTION,
	    CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
	    CDIALOG_BTNFLAG_NO
	);
	g_free(s);
	CDialogSetTransientFor(NULL);
	if(response != CDIALOG_RESPONSE_YES)
	    return;


	/* Remove each selected frame */
	gtk_clist_freeze(clist);
	while(clist->selection != NULL)
	{
	    i = (gint)clist->selection->data;

	    /* Remove the ImgView image frame */
	    ImgViewImageDeleteFrame(img, i);

	    /* Remove the corresponding GtkCList row */
	    gtk_clist_remove(clist, i);
	}
	gtk_clist_thaw(clist);

	has_changes = TRUE;

	if(ImgViewIsPlaying(iv))
	{
	    if(ImgViewGetTotalFrames(iv) <= 1)
		ImgViewPause(iv);
	}
	else
	    ImgViewQueueDrawView(iv);
	update(d);
}

/*
 *	Shift up
 */
static void shift_frame_up(Dialog *d)
{
	gint row, nrows;
	GList *selection = NULL, *glist, *frame_glist1, *frame_glist2;
	GtkCList *clist;
	imgview_struct *iv;
	imgview_image_struct *img;

	if(d == NULL)
	    return;

	iv = d->iv;
	clist = GTK_CLIST(d->clist);
	if(iv == NULL)
	    return;

	img = ImgViewGetImage(iv);
	if(img == NULL)
	    return;

	/* Get copy of selection list, ordered from first to last */
	nrows = clist->rows;
	for(row = 0; row < nrows; row++)
	{
	    if(g_list_find(clist->selection, (gpointer)row) != NULL)
		selection = g_list_append(
		    selection,
		    (gpointer)row
		);
	}

	/* Shift rows up */
	gtk_clist_freeze(clist);
	for(glist = selection; glist != NULL; glist = g_list_next(glist))
	{
	    row = (gint)glist->data;
	    if(row <= 0)
		break;

	    frame_glist1 = g_list_nth(img->frames_list, row);
	    frame_glist2 = g_list_nth(img->frames_list, row - 1);
	    if((frame_glist1 != NULL) && (frame_glist2 != NULL))
	    {
		imgview_frame_struct *frame = IMGVIEW_FRAME(
		    frame_glist1->data
		);
		frame_glist1->data = frame_glist2->data;
		frame_glist2->data = frame;
	    }

	    gtk_clist_swap_rows(clist, row, row - 1);
	}
	gtk_clist_thaw(clist);

	/* Delete copy of selection list */
	g_list_free(selection);

	if(!ImgViewIsPlaying(iv))
	    ImgViewQueueDrawView(iv);
	update(d);
}

/*
 *	Shift down.
 */
static void shift_frame_down(Dialog *d)
{
	gint row, nrows;
	GList *selection = NULL, *glist, *frame_glist1, *frame_glist2;
	GtkCList *clist;
	imgview_struct *iv;
	imgview_image_struct *img;

	if(d == NULL)
	    return;

	iv = d->iv;
	clist = GTK_CLIST(d->clist);
	if(iv == NULL)
	    return;

	img = ImgViewGetImage(iv);
	if(img == NULL)
	    return;

	/* Get copy of selection list, ordered from last to first */
	nrows = clist->rows;
	for(row = nrows - 1; row >= 0; row--)
	{
	    if(g_list_find(clist->selection, (gpointer)row) != NULL)
		selection = g_list_append(
		    selection,
		    (gpointer)row
		);
	}

	/* Shift rows down */
	gtk_clist_freeze(clist);
	for(glist = selection; glist != NULL; glist = g_list_next(glist))
	{
	    row = (gint)glist->data;
	    if(row >= (nrows - 1))
		break;

	    frame_glist1 = g_list_nth(img->frames_list, row);
	    frame_glist2 = g_list_nth(img->frames_list, row + 1);
	    if((frame_glist1 != NULL) && (frame_glist2 != NULL))
	    {
		imgview_frame_struct *frame = IMGVIEW_FRAME(
		    frame_glist1->data
		);
		frame_glist1->data = frame_glist2->data;
		frame_glist2->data = frame;
	    }

	    gtk_clist_swap_rows(clist, row, row + 1);
	}
	gtk_clist_thaw(clist);

	/* Delete copy of selection list */
	g_list_free(selection);

	if(!ImgViewIsPlaying(iv))
	    ImgViewQueueDrawView(iv);
	update(d);
}


/*
 *	Updates the Dialog.
 */
static void update(Dialog *d)
{
	gboolean sensitive;
	gint row, nframes;
	GList *glist;
	GtkCList *clist;
	imgview_struct *iv;

	if(d == NULL)
	    return;

	if(d->freeze_count > 0)
	    return;

	clist = GTK_CLIST(d->clist);
	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;
	iv = d->iv;
	nframes = ImgViewGetTotalFrames(iv);

	d->freeze_count++;

	sensitive = TRUE;
	GTK_WIDGET_SET_SENSITIVE(d->add_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(d->add_mi, sensitive);
	sensitive = (row > -1) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(d->edit_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(d->edit_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(d->remove_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(d->remove_mi, sensitive);

	sensitive = ((row > -1) && (nframes > 1)) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(d->shift_up_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(d->shift_up_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(d->shift_down_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(d->shift_down_mi, sensitive);

	sensitive = (!ImgViewIsPlaying(iv) && (nframes > 1)) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(d->play_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(d->play_mi, sensitive);
	sensitive = ImgViewIsPlaying(iv);
	GTK_WIDGET_SET_SENSITIVE(d->pause_btn, sensitive); 
	GTK_WIDGET_SET_SENSITIVE(d->pause_mi, sensitive);

	d->freeze_count--;
}


/*
 *	Creates a new frame data.
 */
static FrameData *frame_data_new(
	gulong delay
)
{
	FrameData *fd = FRAME_DATA(g_malloc0(sizeof(FrameData)));
	if(fd == NULL)
	    return(NULL);

	fd->delay = delay;

	return(fd);
}

/*
 *	Deletes the frame data.
 */
static void frame_data_delete(FrameData *fd)
{
	if(fd == NULL)
	    return;

	g_free(fd);
}


/*
 *	Returns the image fit type from the specified string.
 */
static img_fit_type img_fit_type_get_from_string(const gchar *s)
{
	const gchar *list[] = IMG_FIT_STRINGS;
	gint i;

	if(s == NULL)
	    return(IMG_FIT_RESIZE);

	for(i = 0; list[i] != NULL; i++)
	{
	    if(!g_strcasecmp(list[i], s))
		return((img_fit_type)i);
	}

	return(IMG_FIT_RESIZE);
}

/*
 *	Returns a dynamically allocated string describing the
 *	specified image fit type.
 */
static gchar *img_fit_string(img_fit_type img_fit)
{
	const gchar *list[] = IMG_FIT_STRINGS;
	gint i;

	for(i = 0; list[i] != NULL; i++)
	{
	    if(i == (gint)img_fit)
		return(STRDUP(list[i]));
	}

	return(NULL);
}


/*
 *	Builds a standard toplevel dialog and returns the pointers to
 *	the created widgets.
 */
static void build_std_dialog(
	gint width, gint height, const gchar *title,
	GtkWidget *ref_toplevel,
	GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
	GtkAccelGroup **accelgrp_rtn
)
{
	gint	bw = GUI_BUTTON_HLABEL_WIDTH_DEF,
		bh = GUI_BUTTON_HLABEL_HEIGHT_DEF,
		border_major = 5;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent, *parent2;
	GtkWidget *toplevel, *main_vbox;


	/* Create keyboard accelerator group */
	accelgrp = gtk_accel_group_new();
	if(accelgrp_rtn != NULL)
	    *accelgrp_rtn = accelgrp;

	/* Create toplevel */
	toplevel = parent = w = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_set_usize(w, width, height);
	gtk_window_set_title(GTK_WINDOW(w), title);
#ifdef PROG_NAME
	gtk_window_set_wmclass(
	    GTK_WINDOW(w), "dialog", PROG_NAME
	);
#endif
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_animate_48x48_xpm);
	}
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(delete_event_cb), toplevel
	);
	gtk_container_border_width(GTK_CONTAINER(w), 0);
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	if((ref_toplevel != NULL) ?
	    GTK_IS_WINDOW(GTK_OBJECT(ref_toplevel)) : FALSE
	)
	{
#if 0
/* Cannot set modal since the file browser wil not function properly */
	    gtk_window_set_modal(GTK_WINDOW(toplevel), TRUE);
#endif
	    gtk_window_set_transient_for(
		GTK_WINDOW(toplevel), GTK_WINDOW(ref_toplevel)
	    );
	}
	if(toplevel_rtn != NULL)
	    *toplevel_rtn = toplevel;

	main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;
	if(client_vbox_rtn != NULL)
	    *client_vbox_rtn = w;

	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Buttons */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Close Button */
	w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_close_20x20_xpm,
#if defined(PROG_LANGUAGE_SPANISH)
	    "Cierre",
#elif defined(PROG_LANGUAGE_FRENCH)
	    "Fin",
#elif defined(PROG_LANGUAGE_GERMAN)
	    "Nah",
#elif defined(PROG_LANGUAGE_ITALIAN)
	    "Vicino",
#elif defined(PROG_LANGUAGE_DUTCH)
	    "Einde",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
	    "Prximo",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
	    "Nr",
#else
	    "Close",
#endif
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(close_cb), toplevel
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_c, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_c);
	gtk_widget_show(w);
}


/*
 *	Queries user to edit frames.
 *
 *	Always returns TRUE.
 */
gboolean iv_edit_frames(
	GtkWidget *ref_toplevel,
	imgview_struct *iv,
	const gchar *filename
)
{
	const gint border_major = 5;
	gchar *s;
	GtkAccelGroup *accelgrp = NULL;
	GtkWidget *toplevel = NULL, *client_vbox = NULL;
	Dialog *d;

	/* Reset globals */
	has_changes = FALSE;

	/* Create dialog base widgets */
	s = g_strdup_printf(
	    "Edit Frames: %s",
	    STRISEMPTY(filename) ? "Untitled" : filename
	);
	build_std_dialog(
	    400, -1,
	    s,
	    ref_toplevel,
	    &toplevel, &client_vbox, &accelgrp
	);
	g_free(s);
	d = DIALOG(g_malloc0(sizeof(Dialog)));
	d->toplevel = toplevel;
	d->accelgrp = accelgrp;
	d->freeze_count = 0;
	d->iv = iv;
	d->img_fit = IMG_FIT_RESIZE_MAINTAIN_ASPECT;

	/* Create Header Dialog widgets */
	if(client_vbox != NULL)
	{
	    const gint	columns = 2,
			bw = GUI_BUTTON_HLABEL_WIDTH,
			bh = GUI_BUTTON_HLABEL_HEIGHT;
	    const GtkTargetEntry dnd_tar_types[] = {
		{ "text/plain",		0,	DND_INFO_TEXT_PLAIN },
		{ "text/uri-list",	0,	DND_INFO_TEXT_URI_LIST },
		{ "STRING",		0,	DND_INFO_STRING }
	    };
	    gchar **strv;
	    GdkWindow *window = toplevel->window;
	    GtkStyle *style = gtk_widget_get_style(toplevel);
	    GtkWidget	*w,
			*parent = client_vbox,
			*parent2, *parent3;
	    GtkCList *clist;
	    imgview_image_struct *img;


	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent = w;

	    /* Left column */
	    w = gtk_vbox_new(TRUE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;
	    /* Icon */
	    w = gtk_pixmap_new_from_xpm_d(
		window, &style->bg[GTK_STATE_NORMAL],
		(guint8 **)icon_animate_32x32_xpm
	    );
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    /* Right column */
	    w = gtk_vbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	    gtk_widget_show(w);
	    parent2 = w;

	    /* GtkScrolledWindow for the GtkCList */
	    w = gtk_scrolled_window_new(NULL, NULL);
	    gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(w),
		GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC
	    );
	    gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* GtkCList */
	    strv = (gchar **)g_malloc(columns * sizeof(gchar *));
	    strv[0] = "Frame";
	    strv[1] = "Delay";
	    d->clist = w = gtk_clist_new_with_titles(columns, strv);
	    g_free(strv);
	    clist = GTK_CLIST(w);
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(clist_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(clist_event_cb), d
	    );
	    gtk_container_add(GTK_CONTAINER(parent3), w);
	    gtk_widget_realize(w);
	    gtk_widget_set_usize(w, -1, 250);
	    gtk_clist_set_selection_mode(clist, GTK_SELECTION_EXTENDED);
	    gtk_clist_column_titles_passive(clist);
	    gtk_clist_set_column_min_width(
		clist, 0, FRAME_ICON_WIDTH
	    );
	    gtk_clist_set_column_auto_resize(
		clist, 0, FALSE
	    );
	    gtk_clist_set_column_auto_resize(
		clist, 1, TRUE
	    );
	    gtk_clist_set_row_height(clist, FRAME_ICON_HEIGHT);
	    gtk_clist_set_shadow_type(clist, GTK_SHADOW_IN);
	    gtk_signal_connect(
		GTK_OBJECT(w), "select_row",
		GTK_SIGNAL_FUNC(clist_select_row_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "unselect_row",
		GTK_SIGNAL_FUNC(clist_unselect_row_cb), d
	    );
	    GUIDNDSetTar(
		w,
		dnd_tar_types,
		sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
		GDK_ACTION_COPY,		/* Actions */
		GDK_ACTION_COPY,		/* Default action if same */
		GDK_ACTION_COPY,		/* Default action */
		clist_drag_data_received_cb,
		d
	    );
	    gtk_widget_show(w);


	    w = gtk_vbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;

	    w = gtk_vbox_new(FALSE, 0);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Add Button */
	    d->add_btn = w = GUIButtonPixmapLabelH(
		(guint8 **)icon_add_20x20_xpm,
		"Add...", NULL
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_set_usize(w, bw, bh);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(add_cb), d
	    );
	    gtk_accel_group_add(
		accelgrp, GDK_a, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	    );
	    GUIButtonLabelUnderline(w, GDK_a);
	    gtk_widget_show(w);

	    /* Edit Button */
	    d->edit_btn = w = GUIButtonPixmapLabelH(
		(guint8 **)icon_edit_20x20_xpm,
		"Edit...", NULL
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_set_usize(w, bw, bh);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(edit_cb), d
	    );
	    gtk_accel_group_add(
		accelgrp, GDK_e, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	    );
	    GUIButtonLabelUnderline(w, GDK_e);
	    gtk_widget_show(w);

	    /* Remove Button */
	    d->remove_btn = w = GUIButtonPixmapLabelH(
		(guint8 **)icon_remove_20x20_xpm,
		"Remove", NULL
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_set_usize(w, bw, bh);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(remove_cb), d
	    );
	    gtk_accel_group_add(
		accelgrp, GDK_r, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	    );
	    GUIButtonLabelUnderline(w, GDK_r);
	    gtk_widget_show(w);


	    w = gtk_vbox_new(FALSE, 0);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Shift Up Button */
	    d->shift_up_btn = w = GUIButtonArrowLabelH(
		GTK_ARROW_UP, 20, 20, "Shift Up", NULL
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_set_usize(w, bw, bh);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(shift_up_cb), d
	    );
	    gtk_accel_group_add(
		accelgrp, GDK_u, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	    );
	    GUIButtonLabelUnderline(w, GDK_u);
	    gtk_widget_show(w);

	    /* Shift Down Button */
	    d->shift_down_btn = w = GUIButtonArrowLabelH(
		GTK_ARROW_DOWN, 20, 20, "Shift Down", NULL
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_set_usize(w, bw, bh);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(shift_down_cb), d
	    );
	    gtk_accel_group_add(
		accelgrp, GDK_d, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"                             
	    );
	    GUIButtonLabelUnderline(w, GDK_d);
	    gtk_widget_show(w);


	    w = gtk_vbox_new(FALSE, 0);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Play Button */
	    d->play_btn = w = GUIButtonPixmapLabelH(
		(guint8 **)icon_play_20x20_xpm,
		"Play", NULL
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_set_usize(w, bw, bh);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(play_cb), d
	    );
	    gtk_accel_group_add(
		accelgrp, GDK_p, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	    );
	    GUIButtonLabelUnderline(w, GDK_p);
	    gtk_widget_show(w);

	    /* Pause Button */
	    d->pause_btn = w = GUIButtonPixmapLabelH(
		(guint8 **)icon_pause_20x20_xpm,
		"Pause", NULL
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_set_usize(w, bw, bh);
	    gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(pause_cb), d
	    );
	    gtk_accel_group_add(
		accelgrp, GDK_s, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	    );
	    GUIButtonLabelUnderline(w, GDK_s);
	    gtk_widget_show(w);


	    /* Right-click menu */
	    if(TRUE)
	    {
		GtkWidget *menu = (GtkWidget *)GUIMenuCreate();
		guint8 **icon;
		const gchar *label;
		guint accel_key, accel_mods;
		gpointer data = d;
		void (*func_cb)(GtkWidget *w, gpointer);

#define DO_ADD_MENU_ITEM_LABEL	{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp,	\
  icon, label, accel_key, accel_mods, NULL,	\
  data, func_cb					\
 );						\
}
#define DO_ADD_MENU_SEP		{		\
 w = GUIMenuItemCreate(				\
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL,	\
  NULL, NULL, 0, 0, NULL,			\
  NULL, NULL					\
 );						\
}
		d->menu = menu;

		icon = (guint8 **)icon_add_20x20_xpm;
		label = "Add...";
		accel_key = GDK_a;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = add_cb;
		DO_ADD_MENU_ITEM_LABEL
		d->add_mi = w;

		icon = (guint8 **)icon_edit_20x20_xpm;
		label = "Edit...";
		accel_key = GDK_e;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = edit_cb;
		DO_ADD_MENU_ITEM_LABEL
		d->edit_mi = w;

		icon = (guint8 **)icon_remove_20x20_xpm;
		label = "Remove";
		accel_key = GDK_r;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = remove_cb;
		DO_ADD_MENU_ITEM_LABEL
		d->remove_mi = w;

		DO_ADD_MENU_SEP

		icon = NULL;
		label = "Shift Up";
		accel_key = GDK_u;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = shift_up_cb;
		DO_ADD_MENU_ITEM_LABEL
		d->shift_up_mi = w;

		icon = NULL;
		label = "Shift Down";
		accel_key = GDK_d;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = shift_down_cb;
		DO_ADD_MENU_ITEM_LABEL
		d->shift_down_mi = w; 

		DO_ADD_MENU_SEP

		icon = (guint8 **)icon_play_20x20_xpm;
		label = "Play";
		accel_key = GDK_p;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = play_cb;
		DO_ADD_MENU_ITEM_LABEL
		d->play_mi = w;

		icon = (guint8 **)icon_pause_20x20_xpm;
		label = "Pause";
		accel_key = GDK_s;
		accel_mods = GDK_CONTROL_MASK;
		func_cb = pause_cb;
		DO_ADD_MENU_ITEM_LABEL
		d->pause_mi = w;
	    }


	    /* Add frames to the clist */
	    img = ImgViewGetImage(iv);
	    if((img != NULL) ? (img->bpp == 4) : FALSE)
	    {
		GList *glist;
		imgview_frame_struct *frame;

		for(glist = img->frames_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
		    frame = IMGVIEW_FRAME(glist->data);
		    if(frame == NULL)
			continue;

		    clist_row_insert(
			clist, -1,
			frame->buf, img->width, img->height, img->bpl,
			frame->delay
		    );
		}
	    }

	    update(d);
	}


	/* Map toplevel */
	gtk_widget_show_raise(toplevel);


	/* Push main loop to wait for user response */
	gtk_main();


	/* Get values from widgets if user clicked on save */
	if(has_changes)
	{

	}

	gtk_clist_freeze(GTK_CLIST(d->clist));
	gtk_clist_clear(GTK_CLIST(d->clist));
	gtk_clist_thaw(GTK_CLIST(d->clist));

	/* Destroy widgets and related resources */
/*	gtk_window_set_modal(GTK_WINDOW(toplevel), FALSE); */
	gtk_window_set_transient_for(GTK_WINDOW(toplevel), NULL);

	GTK_WIDGET_DESTROY(d->clist);
	GTK_WIDGET_DESTROY(d->menu);
	GTK_WIDGET_DESTROY(d->add_btn);
	GTK_WIDGET_DESTROY(d->edit_btn);
	GTK_WIDGET_DESTROY(d->remove_btn);
	GTK_WIDGET_DESTROY(d->shift_up_btn);
	GTK_WIDGET_DESTROY(d->shift_down_btn);
	GTK_WIDGET_DESTROY(d->play_btn);
	GTK_WIDGET_DESTROY(d->pause_btn);
	GTK_WIDGET_DESTROY(d->toplevel);
	GTK_ACCEL_GROUP_UNREF(d->accelgrp);
	g_free(d);

	/* Return user response */
	return(has_changes);
}
