#if !defined(_WIN32)
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"

#include "guiutils.h"
#include "guirgbimg.h"
#include "fb.h"
#include "imgview.h"
#include "printopts.h"
#include "config.h"

#include "images/icon_print_48x48.xpm"
#include "images/icon_print_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"

#include "images/icon_minus_16x16.xpm"
#include "images/icon_plus_16x16.xpm"
#include "images/icon_brightness_16x16.xpm"
#include "images/icon_contrast_16x16.xpm"


/*
 *	Paper Size:
 */
typedef struct {
	gchar		*name,
			*desc;
	gint		width,
			height;
} paper_size_struct;
#define PAPER_SIZE(p)		((paper_size_struct *)(p))


/* Callbacks */
static void PrintDlgPaperSizeRowDataDeleteCB(gpointer data);
static gint PrintDlgDeleteEventCB(GtkWidget *widget, GdkEvent *event, gpointer data);
static void PrintDlgSelectRowCB(
	GtkWidget *widget, gint row, gint column,
	GdkEventButton *event, gpointer data
);
static void PrintDlgPrintCB(GtkWidget *widget, gpointer data);
static void PrintDlgCancelCB(GtkWidget *widget, gpointer data);
static void PrintDlgToggledCB(GtkWidget *widget, gpointer data);
static void PrintDlgAdjustmentValueChangedCB(
	GtkAdjustment *adj, gpointer data
);
static void PrintDlgDPIChangedCB(GtkWidget *widget, gpointer data);
static void PrintDlgOutputChangedCB(GtkWidget *widget, gpointer data);
static void PrintDlgPrintFileBrowseCB(GtkWidget *widget, gpointer data);
static gint PrintDlgPreviewEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static gint PrintDlgPreviewExposeCB(
	GtkWidget *widget, GdkEventExpose *expose, gpointer data
);


/* Utilities */
static void PrintDlgSyncOutputWidgets(print_dlg_struct *pw);
static print_visual PrintDlgVisual(print_dlg_struct *pw);
static print_orientation PrintDlgOrientation(print_dlg_struct *pw);
static gboolean PrintDlgGetDPI(
	print_dlg_struct *pw,
	gint *dpi_x, gint *dpi_y
);
static gboolean PrintDlgPaperSize(
	print_dlg_struct *pw,
	gint *width, gint *height,
	gboolean apply_dpi
);
static gfloat PrintDlgSourceImageAspect(print_dlg_struct *pw);
static gboolean PrintDlgOutputOffsetW(
	print_dlg_struct *pw, gint *x, gint *y
);

static void PrintDlgRevertImages(
	imgview_image_struct *src_img, imgview_image_struct *tar_img
);
static void PrintDlgRevertPreview(print_dlg_struct *pw);
static void PrintDlgUpdatePreviewVisualV(
	imgview_image_struct *img, const print_values_struct *v
);
static void PrintDlgUpdatePreviewVisual(print_dlg_struct *pw);
static void PrintDlgRecreatePreview(print_dlg_struct *pw);
static void PrintDlgResizePreview(print_dlg_struct *pw);
static void PrintDlgPreviewDraw(print_dlg_struct *pw);

print_dlg_struct *PrintDlgNew(void);
void PrintDlgMap(print_dlg_struct *pw);
void PrintDlgUnmap(print_dlg_struct *pw);
void PrintDlgDelete(print_dlg_struct *pw);

/* Front ends */
imgview_image_struct *PrintDlgCreateImageFromValues(
	imgview_image_struct *img,
	gint frame_num,
	const print_values_struct *v,
	const guint8 *bg_color			/* 4 bytes RGBA */
);

gboolean PrintDlgGetResponse(
	imgview_image_struct *img,		/* Image to print */
	gint frame_num,				/* Frame to print */
	print_values_struct *v,
	const guint8 *bg_color,			/* 4 bytes RGBA */
	GtkWidget *toplevel
);

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0) 
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f) 
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? ((x) * -1) : (x))

#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


#define PRINT_DEF_TITLE			"Print"
#define PRINT_DEF_WIDTH			640
#define PRINT_DEF_HEIGHT		480

/* Nominal size of preview drawing area, the nominal size must be
 * square in shape
 */
#define PRINT_PREVIEW_WIDTH		200
#define PRINT_PREVIEW_HEIGHT		200

#define PRINT_CLIST_ROW_SPACING		20


/* Tooltips */
#define PRINT_TTM_PRINT_COMMAND		"\
Command to run the print client plus any arguments"

#define PRINT_TTM_BRIGHTNESS		"Brightness"
#define PRINT_TTM_CONTRAST		"Contrast"

#define PRINT_TTM_DPI_X			"\
Dots per inch (DPI) resolution of paper media width,\
 value must match with value set on printer driver"

#define PRINT_TTM_DPI_Y			"\
Dots per inch (DPI) resolution of paper media height,\
 value must match with value set on printer driver"

#define PRINT_TTM_PREVIEW		"\
Press and hold Button1 to adjust the position, press and hold SHIFT +\
 Button1 (or Button2) to adjust the size"


/* Local GTK+ main level */
static gint main_block_level;
/* Got positive user response if TRUE, otherwise FALSE for cancel */
static gboolean got_response;


/*
 *	Paper Size list row destroy callback.
 */
static void PrintDlgPaperSizeRowDataDeleteCB(gpointer data)
{
	paper_size_struct *psize = PAPER_SIZE(data);
	if(psize == NULL)
	    return;

	g_free(psize->name);
	g_free(psize->desc);
	g_free(psize);
}

/*
 *	Toplevel GtkWindow "delete_event" signal callback.
 */
static gint PrintDlgDeleteEventCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	PrintDlgCancelCB(widget, data);
	return(TRUE);
}

/*
 *	Paper size GtkCList "select_row" signal callback.
 */
static void PrintDlgSelectRowCB(
	GtkWidget *widget, gint row, gint column,
	GdkEventButton *event, gpointer data
)
{
	GtkCList *clist;
	print_dlg_struct *pw = PRINT_DLG(data);
	if((widget == NULL) || (pw == NULL))
	    return;

	clist = GTK_CLIST(widget);

	if(pw->map_state)
	{
	    /* Scroll to row if not visable */
	    if(gtk_clist_row_is_visible(clist, row) != GTK_VISIBILITY_FULL)
	        gtk_clist_moveto(
		    clist,
		    row, 0,	/* Row, column */
		    0.5f, 0.0f	/* Row, column */
	        );
	}

	/* Is this is the paper size clist? */
	if((GtkWidget *)clist == pw->paper_size_clist)
	{
	    /* Since paper size has changed, we need to update the
	     * preview widget size and recreate the preview images
	     */
	    PrintDlgResizePreview(pw);
	}
}


/*
 *	Print callback.
 */
static void PrintDlgPrintCB(GtkWidget *widget, gpointer data)
{
	print_dlg_struct *pw = PRINT_DLG(data);
	if(pw == NULL)
	    return;

	/* Update response code */
	got_response = TRUE;

	/* Break out of highest GTK+ main loop */
	if(main_block_level > 0)
	{
	    main_block_level--;
	    gtk_main_quit();
	}
}

/*
 *	Cancel callback.
 */
static void PrintDlgCancelCB(GtkWidget *widget, gpointer data)
{
	print_dlg_struct *pw = PRINT_DLG(data);
	if(pw == NULL)
	    return;

	/* Update response code */
	got_response = FALSE;

	/* Break out of highest GTK+ main loop */
	if(main_block_level > 0)
	{
	    main_block_level--;
	    gtk_main_quit();
	}
}


/*
 *	All purpose "toggled" signal callback for all radio and
 *	toggle buttons.
 */
static void PrintDlgToggledCB(GtkWidget *widget, gpointer data)
{
	print_dlg_struct *pw = PRINT_DLG(data);
	if((widget == NULL) || (pw == NULL))
	    return;

	if(pw->freeze_count > 0)
	    return;

	/* Only handle toggled to TRUE signals */
	if(GTK_IS_TOGGLE_BUTTON(widget) ?
	    !GTK_TOGGLE_BUTTON_GET_ACTIVE(widget) : TRUE
	)
	    return;

	pw->freeze_count++;

	/* Output to printer? */
	if(widget == pw->output_printer_radio)
	{
	    GTK_WIDGET_SET_SENSITIVE(
		pw->output_printer_entry, TRUE
	    )
	    GTK_WIDGET_SET_SENSITIVE(
		pw->output_file_entry, FALSE
	    )
	    GTK_WIDGET_SET_SENSITIVE(
		pw->output_file_browse_btn, FALSE
	    )
	}
	/* Output to file? */
	else if(widget == pw->output_file_radio)
	{
	    GTK_WIDGET_SET_SENSITIVE(
		pw->output_printer_entry, FALSE
	    )
	    GTK_WIDGET_SET_SENSITIVE(
		pw->output_file_entry, TRUE
	    )
	    GTK_WIDGET_SET_SENSITIVE(
		pw->output_file_browse_btn, TRUE 
	    )
	}

	/* Visual Black & White? */
	else if(widget == pw->visual_bw_radio)
	{
	    PrintDlgUpdatePreviewVisual(pw);
	    PrintDlgPreviewDraw(pw);
	}
	/* Visual Greyscale? */
	else if(widget == pw->visual_greyscale_radio)
	{
	    PrintDlgUpdatePreviewVisual(pw);
	    PrintDlgPreviewDraw(pw);
	}
	/* Visual Color? */
	else if(widget == pw->visual_color_radio)
	{
	    PrintDlgUpdatePreviewVisual(pw);
	    PrintDlgPreviewDraw(pw);
	}

	/* Orientation Portrait? */
	else if(widget == pw->orientation_portrait_radio)
	{
	    /* Need to swap output size */
	    gint height = pw->output_width;
	    pw->output_width = pw->output_height;
	    pw->output_height = height;

	    PrintDlgSyncOutputWidgets(pw);
	    PrintDlgRecreatePreview(pw);
	    PrintDlgPreviewDraw(pw);
	}
	/* Orientation Landscape? */
	else if(widget == pw->orientation_landscape_radio)
	{
	    /* Need to swap output size */
	    gint height = pw->output_width;
	    pw->output_width = pw->output_height;
	    pw->output_height = height;

	    PrintDlgSyncOutputWidgets(pw);
	    PrintDlgRecreatePreview(pw);
	    PrintDlgPreviewDraw(pw);
	}

	pw->freeze_count--;
}

/*
 *	Adjustment "value_changed" signal callback.
 */
static void PrintDlgAdjustmentValueChangedCB(
	GtkAdjustment *adj, gpointer data
)
{
	print_dlg_struct *pw = PRINT_DLG(data);
	if((adj == NULL) || (pw == NULL))
	    return;

	if(pw->freeze_count > 0)
	    return;

	pw->freeze_count++;

	/* Assume brightness or contrast adjustment change */
	PrintDlgUpdatePreviewVisual(pw);

	/* Redraw */
	if(pw->map_state)
	    PrintDlgPreviewDraw(pw);

	pw->freeze_count--;
}


/*
 *	DPI entry "changed" signal callback.
 */
void PrintDlgDPIChangedCB(GtkWidget *widget, gpointer data)
{
	const gchar *s;
	print_dlg_struct *pw = PRINT_DLG(data);
	if((widget == NULL) || (pw == NULL))
	    return;

	if(pw->freeze_count > 0)
	    return;

	pw->freeze_count++;

	/* Get text from entry widget and check if it contains
	 * data
	 */
	s = gtk_entry_get_text(GTK_ENTRY(widget));
	if((s != NULL) ?
	    ((*s != '\0') && !ISBLANK(*s)) : FALSE
	)
	{
	    /* The DPI has changed so the preview widget needs to be
	     * resized and the preview image needs to be recreated to
	     * match the new size of the preview widget
	     */
	    PrintDlgResizePreview(pw);
	}

	pw->freeze_count--;
}


/*
 *	Output geometry widget "changed" or "toggled" signal callback.
 */
static void PrintDlgOutputChangedCB(GtkWidget *widget, gpointer data)
{
	print_dlg_struct *pw = PRINT_DLG(data);
	if((widget == NULL) || (pw == NULL))
	    return;

	if(pw->freeze_count > 0)
	    return;

	pw->freeze_count++;

	if(widget == pw->output_x_spin)
	{
	    pw->output_x = GTK_ENTRY_GET_VALUEI(widget);
	    PrintDlgPreviewDraw(pw);
	}
	else if(widget == pw->output_y_spin)
	{
	    pw->output_y = GTK_ENTRY_GET_VALUEI(widget);
	    PrintDlgPreviewDraw(pw);
	}
	else if(widget == pw->output_width_spin)
	{
	    pw->output_width = GTK_ENTRY_GET_VALUEI(widget);
	    if(pw->output_width > 0)
	    {
		/* Need to adjust height if maintaining aspect? */
		if(pw->output_maintain_aspect)
		{ 
		    gfloat aspect = PrintDlgSourceImageAspect(pw);
		    if(aspect > 0.0f)
		    {
			if(PrintDlgOrientation(pw) == PRINT_ORIENTATION_LANDSCAPE)
			    pw->output_height = pw->output_width * aspect;
			else
			    pw->output_height = pw->output_width / aspect;
		    }
		    gtk_spin_button_set_value(
			GTK_SPIN_BUTTON(pw->output_height_spin),
			(gfloat)pw->output_height
		    );
		}
		/* Recreate preview image due to resize and redraw */
		PrintDlgRecreatePreview(pw);
		PrintDlgPreviewDraw(pw);
	    }
	}
	else if(widget == pw->output_height_spin)
	{
	    pw->output_height = GTK_ENTRY_GET_VALUEI(widget);
	    if(pw->output_height > 0)
	    {
		/* Need to adjust width if maintaining aspect? */
		if(pw->output_maintain_aspect)
		{
		    gfloat aspect = PrintDlgSourceImageAspect(pw);
		    if(aspect > 0.0f)
		    {
			if(PrintDlgOrientation(pw) == PRINT_ORIENTATION_LANDSCAPE)
			    pw->output_width = pw->output_height / aspect;
			else
			    pw->output_width = pw->output_height * aspect;
		    }
		    gtk_spin_button_set_value(
			GTK_SPIN_BUTTON(pw->output_width_spin),
			(gfloat)pw->output_width
		    );
		}
		/* Recreate preview image due to resize and redraw */
		PrintDlgRecreatePreview(pw);
		PrintDlgPreviewDraw(pw);
	    }
	}
	else if(widget == pw->output_maintain_aspect_check)
	{
	    pw->output_maintain_aspect = GTK_TOGGLE_BUTTON_GET_ACTIVE(widget);
	    if(pw->output_maintain_aspect)
	    {
		gfloat aspect = PrintDlgSourceImageAspect(pw);
		if(aspect > 0.0f)
		{
		    /* Update size values */
		    if(PrintDlgOrientation(pw) == PRINT_ORIENTATION_LANDSCAPE)
			pw->output_height = pw->output_width * aspect;
		    else
			pw->output_height = pw->output_width / aspect;
		    gtk_spin_button_set_value(
			GTK_SPIN_BUTTON(pw->output_width_spin),
			(gfloat)pw->output_width
		    );
		    gtk_spin_button_set_value(
			GTK_SPIN_BUTTON(pw->output_height_spin),
			(gfloat)pw->output_height
		    );
		    /* Recreate preview image due to resize and redraw */
		    PrintDlgRecreatePreview(pw);
		    PrintDlgPreviewDraw(pw);
		}
	    }
	}

	pw->freeze_count--;
}

/*
 *	Print to file prompt browse button callback.
 */
static void PrintDlgPrintFileBrowseCB(GtkWidget *widget, gpointer data)
{
	gboolean response;
	gint npaths, total_ftypes = 0;
	const gchar *last_path;
	gchar **paths_list;
	GtkWidget *w, *toplevel;
	GtkEntry *entry;
	fb_type_struct **ftype = NULL, *ftype_rtn;
	print_dlg_struct *pw = PRINT_DLG(data);
	if(pw == NULL)
	    return;

	w = pw->output_file_entry;
	toplevel = pw->toplevel;
	if(w == NULL)
	    return;

	entry = GTK_ENTRY(w);

	/* Get last path */
	last_path = gtk_entry_get_text(entry);

	/* Set up file extension types list */
	ftype = NULL;
	total_ftypes = 0;
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    FTYPE_EXT_PS, "Adobe ISO PostScript"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    FTYPE_EXT_PS2, "Adobe ISO PostScript 2"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    FTYPE_EXT_PS3, "Adobe ISO PostScript 3"
	);
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "All files"
	);

	/* Unset the print dialog as modal */
	gtk_window_set_modal(GTK_WINDOW(toplevel), FALSE);

	/* Map file browser and query user for response */
	FileBrowserSetTransientFor(toplevel);
	response = FileBrowserGetResponse(
#if defined(PROG_LANGUAGE_SPANISH)
            "Escoja El PostScript File",
            "Escoja", "Cancele",
#elif defined(PROG_LANGUAGE_FRENCH)
            "Privilgi PostScript File",
            "Privilgi", "Annuler",
#elif defined(PROG_LANGUAGE_GERMAN)
            "Wahlen Sie PostScript File",
            "Wahlen", "Heben",
#elif defined(PROG_LANGUAGE_ITALIAN)
            "Scegliere PostScript File",
            "Sceglier", "Annullar",
#elif defined(PROG_LANGUAGE_DUTCH) 
            "Selecteer PostScript File",
            "Selecter", "Annuler",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
            "Selecione PostScript File",
            "Selecion", "Cancela",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
            "Velg Ut PostScript File",
            "Velg", "Kanseler",
#else
            "Select PostScript File",
            "Select", "Cancel",
#endif
	    last_path,
	    ftype, total_ftypes,
	    &paths_list, &npaths,
	    &ftype_rtn
	);
	FileBrowserSetTransientFor(NULL);

	/* Set the print dialog as modal */
	gtk_window_set_modal(GTK_WINDOW(toplevel), TRUE);

	if(response)
	{
	    const gchar *path = (npaths > 0) ? paths_list[0] : NULL;
	    if(path != NULL)
		gtk_entry_set_text(entry, path);
	}

	/* Delete the file types list */
	FileBrowserDeleteTypeList(ftype, total_ftypes);
}


/*
 *	Print window's preview_da event handler.
 */
static gint PrintDlgPreviewEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint etype, status = FALSE;
	gboolean need_grab_pointer = FALSE;
	GdkCursor *cur = NULL;
	GdkWindow *window;
	GdkEventConfigure *configure;
	GdkEventCrossing *crossing;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	gboolean keystate;
	guint keyval;
	imgview_image_struct *preview_img;
	print_dlg_struct *pw = PRINT_DLG(data);
	if((widget == NULL) || (event == NULL) || (pw == NULL))
	    return(status);

	/* Get view widget GdkWindow */
	if(GTK_WIDGET_NO_WINDOW(widget))
	    window = NULL;
	else
	    window = widget->window;

	/* Get preview image (may be NULL) */
	preview_img = pw->preview_img;

	/* Get event type */
	etype = *(gint *)event;

	/* Handle by event type */
	switch(etype)
	{
	  case GDK_CONFIGURE:
	    configure = (GdkEventConfigure *)event;
	    PrintDlgRecreatePreview(pw);
	    if(pw->map_state)
		PrintDlgPreviewDraw(pw);
	    status = TRUE;
	    break;

	  case GDK_ENTER_NOTIFY:
	    crossing = (GdkEventCrossing *)event;
	    if(!crossing->send_event)
	    {
#if 0
		/* Need to reset all modifiers */
		if(pw->modifiers)
		{
		    pw->modifiers = 0;
		    if(window != NULL)
			gdk_window_set_cursor(window, NULL);
		}
#endif
		/* Set up widget to grab focus so "key_press_event" and
		 * and "key_release_event" events will be sent to it
		 */
		gtk_widget_grab_focus(widget);

		/* Force tooltip to show, this will send a synthetic
		 * GDK_ENTER_NOTIFY event which we will ignore
		 */
		GUIShowTipsNow(widget);
		status = TRUE;
	    }
	    break;

	  case GDK_LEAVE_NOTIFY:
	    crossing = (GdkEventCrossing *)event;
	    status = TRUE;
	    break;

	  case GDK_KEY_PRESS: case GDK_KEY_RELEASE:
	    key = (GdkEventKey *)event;
	    keyval = key->keyval;
	    keystate = ((etype == GDK_KEY_PRESS) ? TRUE : FALSE);
	    /* Handle by keyval */
	    /* ALT? */
	    if((keyval == GDK_Alt_L) || (keyval == GDK_Alt_R))
	    {
		if(keystate)
		    pw->modifiers |= PRINT_MODIFIER_ALT;
		else
		    pw->modifiers &= ~PRINT_MODIFIER_ALT;
		status = TRUE;
	    }
	    /* CTRL? */
	    else if((keyval == GDK_Control_L) || (keyval == GDK_Control_R))
	    {
		if(keystate)
		    pw->modifiers |= PRINT_MODIFIER_CTRL;
		else
		    pw->modifiers &= ~PRINT_MODIFIER_CTRL;
		status = TRUE;
	    }
	    /* SHIFT? */
	    else if((keyval == GDK_Shift_L) || (keyval == GDK_Shift_R))
	    {
		if(keystate)
		    pw->modifiers |= PRINT_MODIFIER_SHIFT;
		else
		    pw->modifiers &= ~PRINT_MODIFIER_SHIFT;
		status = TRUE;
	    }
	    if(status)
	    {
		if(pw->drag_mode == PRINT_DRAG_NONE)
		{
		    if(pw->modifiers & PRINT_MODIFIER_SHIFT)
		    {
			cur = pw->size_cur;
		    }

		    /* Update cursor */
		    if(window != NULL)
			gdk_window_set_cursor(window, cur);
		}
	    }
	    break;

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    /* Handle by which button number was pressed */
	    switch(button->button)
	    {
	      case GDK_BUTTON1:
		if(pw->modifiers & (PRINT_MODIFIER_ALT | PRINT_MODIFIER_SHIFT))
		{
		    /* Size, same as if GDK_BUTTON2 was pressed */
		    cur = pw->size_cur;
		    need_grab_pointer = TRUE;
		    pw->drag_mode = PRINT_DRAG_SIZE;
		    pw->drag_last_x = (gint)button->x;
		    pw->drag_last_y = (gint)button->y;

		    if(!PrintDlgOutputOffsetW(
			pw, &pw->drag_outline_x[0], &pw->drag_outline_y[0]
		    ))
		    {
			pw->drag_outline_x[0] = 0;
			pw->drag_outline_y[0] = 0;
		    }
		    if(preview_img != NULL)
		    {
			pw->drag_outline_x[1] = pw->drag_outline_x[0] +
			    preview_img->width;
			pw->drag_outline_y[1] = pw->drag_outline_y[0] +
			    preview_img->height;
		    }
		    /* If mapped, then draw only the outline */
		    if(pw->map_state)
		    {
			pw->have_prev_outline = FALSE;
			pw->only_draw_outline = TRUE;;
			PrintDlgPreviewDraw(pw);
		    }
		}
		else
		{
		    /* No keyboard modifiers held, then just translate */
		    cur = pw->translate_cur;
		    need_grab_pointer = TRUE;
		    pw->drag_mode = PRINT_DRAG_TRANSLATE;
		    pw->drag_last_x = (gint)button->x;
		    pw->drag_last_y = (gint)button->y;

		    if(!PrintDlgOutputOffsetW(
		        pw, &pw->drag_outline_x[0], &pw->drag_outline_y[0]
		    ))
		    {
		        pw->drag_outline_x[0] = 0;
		        pw->drag_outline_y[0] = 0;
		    }
		    if(preview_img != NULL)
		    {
		        pw->drag_outline_x[1] = pw->drag_outline_x[0] +
			    preview_img->width;
			pw->drag_outline_y[1] = pw->drag_outline_y[0] +
			    preview_img->height;
		    }
		    /* If mapped, then draw only the outline */
		    if(pw->map_state)
		    {
		        pw->have_prev_outline = FALSE;
		        pw->only_draw_outline = TRUE;;
			PrintDlgPreviewDraw(pw);
		    }
		}
		status = TRUE;
		break;

	      case GDK_BUTTON2:
		/* Size */
		cur = pw->size_cur;
		need_grab_pointer = TRUE;
		pw->drag_mode = PRINT_DRAG_SIZE;
		pw->drag_last_x = (gint)button->x;
		pw->drag_last_y = (gint)button->y;

		if(!PrintDlgOutputOffsetW(
		    pw, &pw->drag_outline_x[0], &pw->drag_outline_y[0]
		))
		{
		    pw->drag_outline_x[0] = 0;
		    pw->drag_outline_y[0] = 0;
		}
		if(preview_img != NULL)
		{
		    pw->drag_outline_x[1] = pw->drag_outline_x[0] +
			preview_img->width;
		    pw->drag_outline_y[1] = pw->drag_outline_y[0] +
			preview_img->height;
		}
		/* If mapped, then draw only the outline */
		if(pw->map_state)
		{
		    pw->have_prev_outline = FALSE;
		    pw->only_draw_outline = TRUE;;
		    PrintDlgPreviewDraw(pw);
		}
		status = TRUE;
		break;
	    }
	    /* Set window cursor if specified */
	    if((window != NULL) && (cur != NULL))
		gdk_window_set_cursor(window, cur);

	    /* Grab pointer? */
	    if(need_grab_pointer && (window != NULL))
	    {
		gdk_pointer_grab(
		    window,
		    FALSE,
		    GDK_BUTTON_PRESS_MASK |
		    GDK_BUTTON_RELEASE_MASK |
		    GDK_POINTER_MOTION_MASK,
		    NULL,
		    GDK_NONE,
		    button->time
		);
		need_grab_pointer = FALSE;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    button = (GdkEventButton *)event;

	    /* Ungrab pointer */
	    if(gdk_pointer_is_grabbed())
		gdk_pointer_ungrab(button->time);

	    /* Handle by current drag mode */
	    switch(pw->drag_mode)
	    {
	      case PRINT_DRAG_TRANSLATE:
		if(TRUE)
		{
		    gint	ox = MIN(
			pw->drag_outline_x[0], pw->drag_outline_x[1]
		    ),
				oy = MIN(
			pw->drag_outline_y[0], pw->drag_outline_y[1]
		    );
		    gint paper_width, paper_height, ww, wh;
		    gfloat wtoo_coeff;


		    ww = widget->allocation.width;
		    wh = widget->allocation.height;

		    if(!PrintDlgPaperSize(pw, &paper_width, &paper_height, TRUE))
		    {
			paper_width = 320;
			paper_height = 240;
		    }

		    if(ww > 0)
			wtoo_coeff = (gfloat)paper_width / (gfloat)ww;
		    else
			wtoo_coeff = 0.0f;
		    pw->output_x = ox * wtoo_coeff;

		    if(wh > 0)
			wtoo_coeff = (gfloat)paper_height / (gfloat)wh;
		    else
			wtoo_coeff = 0.0f;
		    pw->output_y = oy * wtoo_coeff;

		    /* Update output entry widgets to reflect new position */
		    PrintDlgSyncOutputWidgets(pw);
		}
		/* Reset drag mode and redraw */
		pw->drag_mode = PRINT_DRAG_NONE;
		if(pw->map_state)
		    PrintDlgPreviewDraw(pw);
		status = TRUE;
		break;

	      case PRINT_DRAG_SIZE:
		if(TRUE)
		{
		    gint	ow = ABSOLUTE(
			pw->drag_outline_x[1] - pw->drag_outline_x[0]
		    ),
				oh = ABSOLUTE(
			pw->drag_outline_y[1] - pw->drag_outline_y[0]
		    );
		    gint	ww = widget->allocation.width,
				wh = widget->allocation.height;

		    gint paper_width, paper_height;
		    gfloat wtoo_coeff;

		    /* Get paper size */
		    if(!PrintDlgPaperSize(
			pw, &paper_width, &paper_height, TRUE
		    ))
		    {
			paper_width = 320;
			paper_height = 240;
		    }

		    /* Calculate output size */
		    wtoo_coeff = (ww > 0) ?
			((gfloat)paper_width / (gfloat)ww) : 0.0f;
		    if((ow > 0) && (wtoo_coeff > 0.0f))
			pw->output_width = ow * wtoo_coeff;

		    wtoo_coeff = (wh > 0) ?
			((gfloat)paper_height / (gfloat)wh) : 0.0f;
		    if((oh > 0) && (wtoo_coeff > 0.0f))
			pw->output_height = oh * wtoo_coeff;

		    /* Update output entry widgets to reflect new size */
		    PrintDlgSyncOutputWidgets(pw);

		    /* Need to recreate the preview image since it changed
		     * size
		     */
		    PrintDlgRecreatePreview(pw);
		}
		/* Reset drag mode and redraw */
		pw->drag_mode = PRINT_DRAG_NONE;
		if(pw->map_state)
		    PrintDlgPreviewDraw(pw);
		status = TRUE;
		break;

	      default:
		/* Always reset drag mode on button release */
		pw->drag_mode = PRINT_DRAG_NONE;
		break;
	    }
	    /* If no keyboard modifiers are being held then reset cursor */
	    if(!pw->modifiers)
	    {
		if(window != NULL)
		    gdk_window_set_cursor(window, NULL);
	    }
	    break;

	  case GDK_MOTION_NOTIFY:
	    motion = (GdkEventMotion *)event;
	    if(pw->drag_mode != PRINT_DRAG_NONE)
	    {
		gint x, y, dx, dy;
		GdkModifierType mask;

		if(motion->is_hint)
		{
		    gdk_window_get_pointer(
			motion->window, &x, &y, &mask
		    );
		}
		else
		{
		    x = (gint)motion->x;
		    y = (gint)motion->y;
		    mask = motion->state;
		}

		/* Handle by the drag mode */
		switch(pw->drag_mode)
		{
		   case PRINT_DRAG_TRANSLATE:
		    /* Calculate delta values */
		    dx = x - pw->drag_last_x;
		    dy = y - pw->drag_last_y;

		    /* Apply delta values to translate drag outline */
		    pw->drag_outline_x[0] += dx;
		    pw->drag_outline_y[0] += dy;
		    pw->drag_outline_x[1] += dx;
		    pw->drag_outline_y[1] += dy;

		    /* Draw only the outline if mapped */
		    if(pw->map_state)
		    {
		        pw->only_draw_outline = TRUE;
			PrintDlgPreviewDraw(pw);
		    }

		    /* Update last drag positions */
		    pw->drag_last_x = x;
		    pw->drag_last_y = y;

		    status = TRUE;
		    break;

		  case PRINT_DRAG_SIZE:
		    /* Calculate delta values */
		    dx = x - pw->drag_last_x;
		    dy = y - pw->drag_last_y;

		    /* Apply delta values to resize drag outline based
		     * on maintaining aspect or not
		     */
		    if(pw->output_maintain_aspect)
		    {
			/* Maintaining aspect, so get aspect and resize
			 * drag output while maintaining aspect
			 */
			gfloat aspect = PrintDlgSourceImageAspect(pw);
			print_orientation orientation = PrintDlgOrientation(pw);
			gint len;

			if(aspect <= 0.0f)
			    aspect = 1.0f;

			/* Swap aspect if orientation is set to landscape */
			if(orientation == PRINT_ORIENTATION_LANDSCAPE)
			    aspect = 1 / aspect;

			/* Use greater of the two compoent movements */
			if(ABSOLUTE(dx) > ABSOLUTE(dy))
			{
			    pw->drag_outline_x[1] += dx;
			    len = MAX(
				pw->drag_outline_x[1] - 
				pw->drag_outline_x[0], 0
			    );
			    pw->drag_outline_y[1] = pw->drag_outline_y[0] +
				(len / aspect);
			}
			else
			{
			    pw->drag_outline_y[1] += dy;
			    len = MAX(
				pw->drag_outline_y[1] - 
				pw->drag_outline_y[0], 0 
			    );
			    pw->drag_outline_x[1] = pw->drag_outline_x[0] +
				(len * aspect);
			}
		    }
		    else
		    {
			pw->drag_outline_x[1] += dx;
			pw->drag_outline_y[1] += dy;
		    }
		    /* Make sure drag outline does not get smaller than
		     * its origin
		     */
		    if(pw->drag_outline_x[1] < pw->drag_outline_x[0])
			pw->drag_outline_x[1] = pw->drag_outline_x[0];
		    if(pw->drag_outline_y[1] < pw->drag_outline_y[0])
			pw->drag_outline_y[1] = pw->drag_outline_y[0];

		    /* Draw only the outline if mapped */
		    if(pw->map_state)
		    {
			pw->only_draw_outline = TRUE;
			PrintDlgPreviewDraw(pw);
		    }

		    /* Update last drag positions */
		    pw->drag_last_x = x;
		    pw->drag_last_y = y;
		    status = TRUE;
		    break;
		}
	    }
	    break;
	}

	return(status);
}

/*
 *	Print window's preview_da widget "expose_event" callback.
 */
static gint PrintDlgPreviewExposeCB(
	GtkWidget *widget, GdkEventExpose *expose, gpointer data
)
{
	GdkGC *gc_base, *gc_invert;
	gboolean only_draw_outline;
	GdkWindow *window;
	GdkDrawable *d;
	GtkWidget *w;
	gint ww, wh;
	imgview_frame_struct *frame;
	imgview_image_struct *img;
	print_dlg_struct *pw = PRINT_DLG(data);
	if((widget == NULL) || (pw == NULL))
	    return(FALSE);

	/* Get marker for if we only want to draw the drag outlines */
	only_draw_outline = pw->only_draw_outline;
	pw->only_draw_outline = FALSE;	/* Always reset it in this function */

	/* Get preview drawing area widget */
	w = pw->preview_da;
	if(w != widget)
	    return(FALSE);

	ww = w->allocation.width;
	wh = w->allocation.height;
	if((ww <= 0) || (wh <= 0))
	    return(FALSE);

	/* Get pointers to the GCs */
	gc_base = pw->gc_base;
	gc_invert = pw->gc_invert;

	/* Must have the base GC */
	if(gc_base == NULL)
	    return(FALSE);

	/* Get preview drawing area widget's window */
	if(GTK_WIDGET_NO_WINDOW(w))
	    window = NULL;
	else
	    window = w->window;
	if(window == NULL)
	    return(FALSE);

	d = (GdkDrawable *)window;


	/* Clear background as needed */
	if(!only_draw_outline)
	{
	    gdk_draw_rectangle(d, gc_base, TRUE, 0, 0, ww, wh);
	}

	/* Get preview image and check if it contains data? */
	img = pw->preview_img;
	frame = ImgViewImageGetFrame(img, 0);
	if((frame != NULL) ?
	    ((frame->buf != NULL) && !only_draw_outline) : FALSE
	)
	{
	    gint x, y;

	    if(!PrintDlgOutputOffsetW(pw, &x, &y))
	    {
		x = 0;
		y = 0;
	    }

	    switch(img->bpp)
	    {
	      case 4:
		gdk_draw_rgb_32_image(
		    d, gc_base,
		    x, y,
		    img->width, img->height,
		    GDK_RGB_DITHER_NORMAL,
		    (guchar *)frame->buf,
		    img->bpl
		);
		break;

	      case 3:
		gdk_draw_rgb_image(
		    d, gc_base,
		    x, y,
		    img->width, img->height,
		    GDK_RGB_DITHER_NORMAL,
		    (guchar *)frame->buf,
		    img->bpl
		);
		break;
	    }
	}


	/* Clear previous drag marks if defined and if only drawing
	 * for drag outlines.
	 */
	if(pw->have_prev_outline && only_draw_outline && (gc_invert != NULL))
	{
	     gint        x = MIN(pw->prev_outline_x[0], pw->prev_outline_x[1]),
			 y = MIN(pw->prev_outline_y[0], pw->prev_outline_y[1]);
	     gint        width = ABSOLUTE(
		pw->prev_outline_x[1] - pw->prev_outline_x[0]
			 ),
			 height = ABSOLUTE(
		pw->prev_outline_y[1] - pw->prev_outline_y[0]
			 );

	    if((width > 0) && (height > 0))
		gdk_draw_rectangle(
		    d, gc_invert, FALSE, x, y, width, height
		);

	    pw->have_prev_outline = FALSE;
	}

	/* Draw new drag marks? */
	if((pw->drag_mode != PRINT_DRAG_NONE) && (gc_invert != NULL))
	{
	    gint	x = MIN(pw->drag_outline_x[0], pw->drag_outline_x[1]),
			y = MIN(pw->drag_outline_y[0], pw->drag_outline_y[1]);
	    gint	width = ABSOLUTE(
		pw->drag_outline_x[1] - pw->drag_outline_x[0]
			),
			height = ABSOLUTE(
		pw->drag_outline_y[1] - pw->drag_outline_y[0]
			);

	    if((width > 0) && (height > 0))
	    {
		gdk_draw_rectangle(
		    d, gc_invert, FALSE, x, y, width, height
		);

		/* Record newly drawn drag outline marks */
		pw->have_prev_outline = TRUE;
		pw->prev_outline_x[0] = pw->drag_outline_x[0];
		pw->prev_outline_y[0] = pw->drag_outline_y[0];
		pw->prev_outline_x[1] = pw->drag_outline_x[1];
		pw->prev_outline_y[1] = pw->drag_outline_y[1];
	    }
	}

	return(TRUE);
}


/*
 *	Sets the Print Window's output widget values to match the current
 *	output geometry values.
 */
static void PrintDlgSyncOutputWidgets(print_dlg_struct *pw)
{
	GtkWidget *w;

	if(pw == NULL)
	    return;

	pw->freeze_count++;

	/* Begin updating output entry widgets */
	w = pw->output_x_spin;
	if(w != NULL)
	    gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(w), (gfloat)pw->output_x
	    );
	w = pw->output_y_spin;
	if(w != NULL)
	    gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(w), (gfloat)pw->output_y
	    );
	w = pw->output_width_spin;
	if(w != NULL)
	    gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(w), (gfloat)pw->output_width
	    );
	w = pw->output_height_spin;
	if(w != NULL)
	    gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(w), (gfloat)pw->output_height
	    );
	w = pw->output_maintain_aspect_check;
	if(w != NULL)
	    gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(w), pw->output_maintain_aspect
	    );

	pw->freeze_count--;
}

/*
 *	Gets the Print Window's print visual.
 */
static print_visual PrintDlgVisual(print_dlg_struct *pw)
{
	if(pw == NULL)
	    return(PRINT_VISUAL_BW);

	if(GTK_TOGGLE_BUTTON_GET_ACTIVE(
	    pw->visual_bw_radio
	))
	    return(PRINT_VISUAL_BW);
	else if(GTK_TOGGLE_BUTTON_GET_ACTIVE(
	    pw->visual_greyscale_radio
	))
	    return(PRINT_VISUAL_GREYSCALE);
	else if(GTK_TOGGLE_BUTTON_GET_ACTIVE(
	    pw->visual_color_radio
	))
	    return(PRINT_VISUAL_RGB);
	else
	    return(PRINT_VISUAL_BW);
}

/*
 *	Gets the Print Window's print orientation.
 */
static print_orientation PrintDlgOrientation(print_dlg_struct *pw)
{
	GtkWidget *w = (pw != NULL) ?
	    pw->orientation_portrait_radio : NULL;
	if(w == NULL)
	    return(PRINT_ORIENTATION_PORTRAIT);

	return(GTK_TOGGLE_BUTTON_GET_ACTIVE(w) ?
	    PRINT_ORIENTATION_PORTRAIT : PRINT_ORIENTATION_LANDSCAPE
	);
}

/*
 *	Gets teh DPI (dots per inch) value, returns TRUE if the value
 *	was obtained or FALSE on failure.
 */
static gboolean PrintDlgGetDPI(
	print_dlg_struct *pw,
	gint *dpi_x, gint *dpi_y
)
{
	GtkWidget *w;

	if(dpi_x != NULL)
	    (*dpi_x) = -1;
	if(dpi_y != NULL)
	    (*dpi_y) = -1;

	if(pw == NULL)
	    return(FALSE);


	w = pw->dpi_x_entry;
	if((w != NULL) ? GTK_IS_ENTRY(w) : 0)
	{
	    const gchar *s = gtk_entry_get_text(
		GTK_ENTRY(w)
	    );
	    if((s != NULL) && (dpi_x != NULL))
		(*dpi_x) = atoi(s);
	}

	w = pw->dpi_y_entry;
	if((w != NULL) ? GTK_IS_ENTRY(w) : 0)
	{
	    const gchar *s = gtk_entry_get_text(
		GTK_ENTRY(w)
	    );
	    if((s != NULL) && (dpi_y != NULL))
		(*dpi_y) = atoi(s);
	}

	return(TRUE);
}

/*
 *	Gets the paper size, returns TRUE if the value was obtained or
 *	FALSE on failure.
 *
 *	If apply_dpi is TRUE then the dpi will be applied to the paper
 *	size before it's returned.
 */
static gboolean PrintDlgPaperSize(
	print_dlg_struct *pw,
	gint *width, gint *height,
	gboolean apply_dpi
)
{
	GList *glist;
	gint row;
	const paper_size_struct *psize;
	gint orig_width, orig_height;
	GtkCList *clist = (pw != NULL) ?
	    (GtkCList *)pw->paper_size_clist : NULL;

	if(width != NULL)
	    *width = -1;
	if(height != NULL)
	    *height = -1;

	if(clist == NULL)
	    return(FALSE);

	/* Get last selected row */
	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;

	/* Get paper size from the selected row */
	if((row >= 0) && (row < clist->rows))
	    psize = PAPER_SIZE(
		gtk_clist_get_row_data(clist, row)
	    );
	else
	    psize = NULL;
	if(psize == NULL)
	    return(FALSE);

	/* Get original paper size */
	orig_width = psize->width;
	orig_height = psize->height;
	if((orig_width <= 0) || (orig_height <= 0))
	    return(FALSE);

	/* Apply DPI to paper size? */
	if(apply_dpi)
	{
	    gint dpi_x, dpi_y;

	    /* Get DPI */
	    if(PrintDlgGetDPI(pw, &dpi_x, &dpi_y))
	    {
		/* Modify width and height to match DPI */
		orig_width = (gfloat)orig_width * (gfloat)dpi_x / 100.0f;
		orig_height = (gfloat)orig_height * (gfloat)dpi_y / 100.0f;
	    }
	}

	/* Update returns */
	if(width != NULL)
	    *width = orig_width;
	if(height != NULL)
	    *height = orig_height;

	return(TRUE);
}

/*
 *	Returns the Print Window's source image aspect.
 */
static gfloat PrintDlgSourceImageAspect(print_dlg_struct *pw)
{
	const imgview_image_struct *img = (pw != NULL) ? pw->src_img : NULL;
	if(img == NULL)
	    return(0.0f);

	if(img->height <= 0)
	    return(0.0f);
	else
	    return((gfloat)img->width / (gfloat)img->height);
}


/*
 *	Returns the Print Window's output image upper-left corner offset from
 *	the window's upper-left corner (regardless of orientation).
 *
 *	Returns TRUE if the output x and y offsets were obtained
 *	or FALSE on error.
 */
static gboolean PrintDlgOutputOffsetW(
	print_dlg_struct *pw, gint *x, gint *y
)
{
	gfloat otow_coeff;
	gint paper_width, paper_height;
	GtkWidget *w = (pw != NULL) ? pw->preview_da : NULL;
	if(w == NULL)
	    return(FALSE);

	if(!PrintDlgPaperSize(
	    pw, &paper_width, &paper_height, TRUE
	))
	    return(FALSE);

	if((paper_width <= 0) || (paper_height <= 0))
	{
	    if(x != NULL)
		*x = 0;
	    if(y != NULL)
		*y = 0;
	    return(TRUE);
	}

	/* Calculate offsets */
	otow_coeff = (gfloat)w->allocation.width / (gfloat)paper_width;
	if(x != NULL)
	    *x = pw->output_x * otow_coeff;
	otow_coeff = (gfloat)w->allocation.height / (gfloat)paper_height;
	if(y != NULL)
	    *y = pw->output_y * otow_coeff;

	return(TRUE);
}


/*
 *	Copies the image data from src_img to tar_img.
 *
 *	Both images must have the same size and depth.
 */
static void PrintDlgRevertImages(
	imgview_image_struct *src_img, imgview_image_struct *tar_img
)
{
	gint bpp, src_bpl, tar_bpl, min_bpl, height;
	const guint8 *src_data, *src_line, *src_line_end;
	guint8 *tar_data, *tar_line;
	imgview_frame_struct *src_frame, *tar_frame;

	if((src_img == NULL) || (tar_img == NULL))
	    return;

	/* Source image geometry and bpp must match the target image */
	if((src_img->bpp != tar_img->bpp) ||
	   (src_img->width != tar_img->width) ||
	   (src_img->height != tar_img->height)
	)
	{
	    g_printerr(
"PrintDlgRevertImages(): BPP or geometry mismatch, cannot copy.\n"
	    );
	    return;
	}

	bpp = src_img->bpp;
	height = src_img->height;

	src_bpl = src_img->bpl;
	tar_bpl = tar_img->bpl;
	min_bpl = src_img->width * bpp;

	src_frame = ImgViewImageGetFrame(src_img, 0);
	tar_frame = ImgViewImageGetFrame(tar_img, 0);

	/* Get pointers to both image's buffers, both must be allocated */
	src_data = src_frame->buf;
	tar_data = tar_frame->buf;
	if((src_data == NULL) || (tar_data == NULL))
	    return;

	for(src_line = src_data,
	    src_line_end = src_line + (src_bpl * height),
	    tar_line = tar_data;
	    src_line < src_line_end;
	    src_line += src_bpl,
	    tar_line += tar_bpl
	)
	    memcpy(tar_line, src_line, min_bpl);
}


/*
 *	Resets the image data in preview_img with the data from 
 *	orig_preview_img.
 */
static void PrintDlgRevertPreview(print_dlg_struct *pw)
{
	if(pw == NULL)
	    return;

	/* Copy data from the original preview image to the preview
	 * image that will be ready for drawing on preview_da
	 */
	PrintDlgRevertImages(
	    pw->orig_preview_img,	/* Source */
	    pw->preview_img		/* Target */
	);
}


/*
 *	Modifies the ImgView image color, brightness, and contrast
 *	with respect to the specified print values.
 */
static void PrintDlgUpdatePreviewVisualV(
	imgview_image_struct *img, const print_values_struct *v
)
{
	guint vflags;
	print_visual visual = PRINT_VISUAL_BW;
	gfloat brightness = 0.0f, contrast = 0.0f;
	gint i, bpp, min_bpp, bpl, width, height;
	guint8 *tar_data, *tar_line, *tar_line_end, *tar_ptr, *tar_end;
	guint32 cur_byte, avg_byte, bw_byte;
	gint brightness_delta;
	imgview_frame_struct *frame;

	if((img == NULL) || (v == NULL))
	    return;

	/* Get image values */
	width = img->width;
	height = img->height;
	bpp = img->bpp;
	bpl = img->bpl;
	frame = ImgViewImageGetFrame(img, 0);
	tar_data = (frame != NULL) ? frame->buf : NULL;
	if(tar_data == NULL)
	    return;

	/* Calculate minimized depth to be 3 bytes per pixel or less */
	min_bpp = MIN(bpp, 3);

	/* Get print values */
	vflags = v->flags;
	if(vflags & PRINT_VALUE_VISUAL)
	    visual = v->visual;
	if(vflags & PRINT_VALUE_BRIGHTNESS)
	    brightness = v->brightness;
	if(vflags & PRINT_VALUE_CONTRAST)
	    contrast = v->contrast;

	/* Calculate brightness delta byte value */
	brightness_delta = (gint)(0xff * brightness);

	/* Iterate through image */
	for(tar_line = tar_data,
	    tar_line_end = tar_line + (bpl * height);
	    tar_line < tar_line_end;
	    tar_line += bpl
	)
	{
	    tar_ptr = tar_line;
	    tar_end = tar_ptr + (width * bpp);

	    while(tar_ptr < tar_end)
	    {
		/* Begin per pixel operations */

		/* Calculate the average value, we will need it in
		 * several places here
		 *
		 * The average byte value is from 0x00 to 0xff
		 */
		avg_byte = (gint)(
		    (0.299 * tar_ptr[0]) +
		    (0.587 * tar_ptr[1]) +
		    (0.114 * tar_ptr[2])
		);

		/* Make pixel grey? */
		switch(visual)
		{
		  case PRINT_VISUAL_BW:
		    bw_byte = (avg_byte >= 0x80) ? 0xff : 0x00;
		    for(i = 0; i < min_bpp; i++)
			tar_ptr[i] = bw_byte;
		    break;
		  case PRINT_VISUAL_GREYSCALE:
		    for(i = 0; i < min_bpp; i++)
			tar_ptr[i] = avg_byte;
		    break;
		  case PRINT_VISUAL_RGB:
		    break;
		}

		/* Adjust brightness? */
		if(brightness_delta != 0)
		{
		    for(i = 0; i < min_bpp; i++)
			tar_ptr[i] = (guint8)CLIP(
			    (gint)tar_ptr[i] + brightness_delta,
			    0x00, 0xff
			);
		}
		/* Adjust contrast? */
		if(contrast != 0.0f)
		{
		    for(i = 0; i < min_bpp; i++)
		    {
			cur_byte = (guint32)(tar_ptr[i]);
			tar_ptr[i] = (guint8)CLIP(
			    (gint)cur_byte +
			    (gint)(((gint)cur_byte - (gint)0x80) * contrast),
			    0x00, 0xff
			);
		    }
		}

		tar_ptr += bpp;
	    }
	}
}


/*
 *	Updates the Print Window's preview_img to reflect the current
 *	visual values.
 *
 *	PrintDlgRevertPreview() will be called to revert the data.
 */
static void PrintDlgUpdatePreviewVisual(print_dlg_struct *pw)
{
	print_values_struct *v;

	if(pw == NULL)
	    return;

	/* Copy data from orig_preview_img to preview_img */
	PrintDlgRevertPreview(pw);

	/* Set up print values to match the print window's current
	 * print values
	 */
	v = PRINT_VALUES(
	    g_malloc0(sizeof(print_values_struct))
	);
	v->flags = PRINT_VALUE_BRIGHTNESS | PRINT_VALUE_CONTRAST |
	    PRINT_VALUE_VISUAL;
	v->brightness = GTK_ADJUSTMENT_GET_VALUE(pw->brightness_adj);
	v->contrast = GTK_ADJUSTMENT_GET_VALUE(pw->contrast_adj);
	v->visual = PrintDlgVisual(pw);

	/* Modify preview_img's color, brightness, and contrast */
	PrintDlgUpdatePreviewVisualV(pw->preview_img, v);

	g_free(v);
}

/*
 *	Recreates the original preview image orig_preview_img and
 *	the preview image preview_img.
 *
 *	The new preview_img will contain the same data as 
 *	orig_preview_img.
 *
 *	This function will call PrintDlgUpdatePreviewVisual() to make sure
 *	the new images are updated and ready to be drawn.
 */
static void PrintDlgRecreatePreview(print_dlg_struct *pw)
{
	print_visual visual;
	print_orientation orientation;
	gint i;
	gint paper_width, paper_height;
	gint output_width, output_height;
	gint new_width, new_height;
	gfloat x_coeff, y_coeff;
	imgview_frame_struct	*src_frame,
				*tar_frame;
	imgview_image_struct	*tar_img,
				*src_img;
	GtkWidget *w = (pw != NULL) ? pw->preview_da : NULL;
 	if(w == NULL)
	    return;

	/* Delete existing preview images */
	ImgViewImageDelete(pw->preview_img);
	pw->preview_img = NULL;
	ImgViewImageDelete(pw->orig_preview_img);
	pw->orig_preview_img = NULL;

	/* Get the source image */
	src_img = pw->src_img;
	src_frame = ImgViewImageGetFrame(src_img, pw->src_frame_num);
	if((src_frame != NULL) ? (src_frame->buf == NULL) : TRUE)
	    return;

	/* Get current print values from the print window */
	orientation = PrintDlgOrientation(pw);
	visual = PrintDlgVisual(pw);
	if(!PrintDlgPaperSize(pw, &paper_width, &paper_height, TRUE))
	    return;

	/* Get output image size in pixels, we can ignore the
	 * orientation because the output size should already be
	 * updated current orientation
	 */
	output_width = pw->output_width;
	output_height = pw->output_height;
	if((output_width <= 0) || (output_height <= 0))
	    return;

	/* Calculate coefficients that represent the amount of
	 * the output image visible on the paper
	 */
	x_coeff = (gfloat)output_width / (gfloat)paper_width;
	y_coeff = (gfloat)output_height / (gfloat)paper_height;

	/* Calculate the size of the new preview image */
	new_width = (gint)(w->allocation.width * x_coeff);
	new_height = (gint)(w->allocation.height * y_coeff);

	if((new_width <= 0) || (new_height <= 0))
	    return;


	/* Create the new preview images */
	pw->orig_preview_img = tar_img = ImgViewImageNew(
	    new_width, new_height, src_img->bpp, 0
	);
	i = ImgViewImageAppendFrameNew(tar_img);
	tar_frame = ImgViewImageGetFrame(tar_img, i);

	pw->preview_img = ImgViewImageNew(
	    new_width, new_height, src_img->bpp, 0
	);
	i = ImgViewImageAppendFrameNew(pw->preview_img);

	if((tar_frame != NULL) ? (tar_frame->buf != NULL) : FALSE)
	{
	    const gint bpp = src_img->bpp;
	    if(bpp != tar_img->bpp)
	    {
		g_printerr(
"PrintDlgRecreatePreview(): Error:\
Source and target images do not have the same bytes per pixel.\n"
		);
		return;
	    }

	    /* Copy/resize by orientation */
	    if(orientation == PRINT_ORIENTATION_LANDSCAPE)
	    {
		/* Create a tempory image rotated 90 degrees of the
		 * target image's size
		 */
		imgview_frame_struct *tmp_frame;
		imgview_image_struct *tmp_img = ImgViewImageNew(
		    tar_img->height, tar_img->width,
		    tar_img->bpp, 0
		);
		i = ImgViewImageAppendFrameNew(tmp_img);
		tmp_frame = ImgViewImageGetFrame(tmp_img, i);
		if(tmp_frame != NULL)
		{
		    /* Copy/resize source image to tempory image */
		    GUIImageBufferResize(
			bpp,
			src_frame->buf,
			src_img->width, src_img->height, src_img->bpl,
			tmp_frame->buf,
			tmp_img->width, tmp_img->height, tmp_img->bpl
		    );

		    /* Copy/rotate tempory image to target image */
		    GUIImageBufferRotateCCW90(
			bpp,
			tmp_frame->buf,
			tmp_img->width, tmp_img->height, tmp_img->bpl,
			tar_frame->buf,
			tar_img->width, tar_img->height, tar_img->bpl
		    );
		}

		/* Delete the tempory image */
		ImgViewImageDelete(tmp_img);
	    }
	    else
	    {
		/* Copy/resize the source image to the target image */
		GUIImageBufferResize(
		    bpp,
		    src_frame->buf,
		    src_img->width, src_img->height, src_img->bpl,
		    tar_frame->buf,
		    tar_img->width, tar_img->height, tar_img->bpl
		);
	    }

	    /* Apply background to the target image */
	    GUIImageBufferFlattenWithBG(
		bpp,
		tar_frame->buf,
		tar_img->width, tar_img->height, tar_img->bpl,
		pw->bg_color
	    );
	}

	/* This will call PrintDlgRevertPreview() to first copy the data
	 * from orig_preview_img to preview_img and modify preview_img to
	 * reflect the current visual settings
	 */
	PrintDlgUpdatePreviewVisual(pw);
}

/*
 *	Resizes the Print Window's preview GtkDrawingArea.
 */
static void PrintDlgResizePreview(print_dlg_struct *pw)
{
	gint width, height;	/* Nominal size of preview_da */
	gint new_width, new_height;
	gint paper_width, paper_height;
	gfloat aspect;
	GtkWidget *w = (pw != NULL) ? pw->preview_da : NULL;
	if(w == NULL)
	    return;

	/* Set nominal display size of preview drawing area */
	width = PRINT_PREVIEW_WIDTH;
	height = PRINT_PREVIEW_HEIGHT;

	/* Get new paper size */
	if(!PrintDlgPaperSize(pw, &paper_width, &paper_height, TRUE))
	    return;
	if(paper_height <= 0)
	    return;

	/* Calculate aspect ratio of paper */
	aspect = (gfloat)paper_width / (gfloat)paper_height;
	if(aspect <= 0.0f)
	    return;

	/* Calculate new size for preview drawing area */
	if(aspect >= 1.0f)
	{
	    /* Width is longer than height */
	    new_width = width;
	    new_height = height / aspect;
	}
	else
	{
	    /* Height is longer than width */
	    new_width = width * aspect;
	    new_height = height;
	}

	/* Set new size of preview drawing area */
	w = pw->preview_da;
	if(w != NULL)
	    gtk_widget_set_usize(w, new_width, new_height);

	/* Notify toplevel about resize */
	w = pw->toplevel;
	if(w != NULL)
	    gtk_widget_queue_resize(w);

	/* We do not need to recreate the preview images since a
	 * "configure_event" will be generated and the signal handler
	 * will see it and recreate the preview images
	 */
}


/*
 *	Redraws the Print Window's preview.
 */
static void PrintDlgPreviewDraw(print_dlg_struct *pw)
{
	if(pw == NULL)
	    return;

	PrintDlgPreviewExposeCB(pw->preview_da, NULL, pw);
}



/*
 *	Creates a new print window.
 */
print_dlg_struct *PrintDlgNew(void)
{
	gint	bw = GUI_BUTTON_HLABEL_WIDTH_DEF,
		bh = GUI_BUTTON_HLABEL_HEIGHT_DEF,
		border_major = 5,
		border_minor = 2;
	gchar *heading[2];
	GSList *gslist;
	GdkColormap *colormap;
	GdkWindow *window;
	GtkStyle *style;
	GtkAccelGroup *accelgrp;
	GtkAdjustment *adj;
	GtkWidget	*w, *parent, *parent2, *parent3, *parent4,
			*parent5, *parent6, *main_vbox, *scroll_parent;
	GtkCList *clist;
	gpointer entry, browse_btn;
	print_dlg_struct *pw = PRINT_DLG(
	    g_malloc0(sizeof(print_dlg_struct))
	);
	if(pw == NULL)
	    return(NULL);

	/* Begin resetting values */
	pw->initialized = TRUE;
	pw->map_state = FALSE;
	pw->freeze_count = 0;
	pw->busy_count = 0;
	pw->src_img = NULL;
	pw->preview_img = NULL;
	pw->orig_preview_img = NULL;
	pw->drag_mode = PRINT_DRAG_NONE;

	/* Load pointer cursors */
	pw->busy_cur = gdk_cursor_new(GDK_WATCH);
	pw->translate_cur = gdk_cursor_new(GDK_FLEUR);
	pw->size_cur = gdk_cursor_new(GDK_SIZING);

	/* Keyboard accelerator group */
	pw->accelgrp = accelgrp = gtk_accel_group_new();

	/* Toplevel */
	gtk_widget_push_visual(gdk_rgb_get_visual());
	gtk_widget_push_colormap(gdk_rgb_get_cmap());
	pw->toplevel = w = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_pop_visual();
	gtk_widget_pop_colormap();
	gtk_window_set_policy(
	    GTK_WINDOW(w), TRUE, TRUE, TRUE
	);
	gtk_widget_set_usize(w, PRINT_DEF_WIDTH, PRINT_DEF_HEIGHT);
	gtk_window_set_title(GTK_WINDOW(w), PRINT_DEF_TITLE);
#ifdef PROG_NAME
	gtk_window_set_wmclass(
	    GTK_WINDOW(w), "dialog", PROG_NAME
	);
#endif
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_print_48x48_xpm);
	}
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(PrintDlgDeleteEventCB), pw
	);
	gtk_container_border_width(GTK_CONTAINER(w), 0);
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	parent = w;


	/* Main vbox */
	main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	/* Vbox for child widgets */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(main_vbox), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent = w;


	/* Output widgets */
	w = gtk_frame_new("Print To");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	/* Hbox for print command widgets */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Print to printer radio */
	gslist = NULL;
	pw->output_printer_radio = w =
	    gtk_radio_button_new_with_label(gslist, "Printer");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgToggledCB),
	    (gpointer)pw
	);
	gtk_widget_set_usize(w, 75, -1);
	gtk_widget_show(w);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	/* Print command prompt */
	w = (GtkWidget *)GUIPromptBar(
	    NULL, "Print Command:",
	    NULL, &entry
	);
	pw->output_printer_entry = (GtkWidget *)entry;
	GUISetWidgetTip((GtkWidget *)entry, PRINT_TTM_PRINT_COMMAND);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);

	/* Hbox for print to file widgets */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;
	/* Radio button */
	pw->output_file_radio = w =
	    gtk_radio_button_new_with_label(gslist, "File");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgToggledCB), pw
	);
	gtk_widget_set_usize(w, 75, -1);
	gtk_widget_show(w);
	/* Prompt with browse */
	w = GUIPromptBarWithBrowse(
	    NULL, "Name:",
	    NULL, &entry, &browse_btn,
	    (gpointer)pw, PrintDlgPrintFileBrowseCB
	);
	pw->output_file_entry = (GtkWidget *)entry;
	pw->output_file_browse_btn = (GtkWidget *)browse_btn;
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);


	/* Hbox to divide into two columns */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;


	/* Left side vbox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;


	/* Visual */
	/* Frame */
	w = gtk_frame_new("Visual");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;
	/* Vbox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;

	/* Vbox for brightness ticks and scale */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Hbox for ticks */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_minus_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_brightness_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_plus_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Brightness scale */
	pw->brightness_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,		/* Initial value */
	    -1.0f,		/* Minimum */
	    1.0f,		/* Maximum */
	    0.1f,		/* Step inc */
	    0.1f,		/* Page inc */
	    0.0f		/* Page size */
	);
	/* Scale */
	gtk_object_ref(GTK_OBJECT(adj));	/* Keep adj around */
	pw->brightness_scale = w = gtk_hscale_new(adj);
	gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	gtk_scale_set_value_pos(GTK_SCALE(w), GTK_POS_RIGHT);
	gtk_scale_set_digits(GTK_SCALE(w), 2);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(PrintDlgAdjustmentValueChangedCB), pw
	);
	GUISetWidgetTip(w, PRINT_TTM_BRIGHTNESS);
	gtk_widget_show(w);

	/* Vbox for contrast ticks and scale */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Hbox for ticks */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_minus_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_contrast_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_plus_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Contrast scale */
	pw->contrast_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,		/* Initial value */
	    -1.0f,		/* Minimum */
	    1.0f,		/* Maximum */
	    0.1f,		/* Step inc */
	    0.1f,		/* Page inc */
	    0.0f		/* Page size */
	);
	/* Scale */
	gtk_object_ref(GTK_OBJECT(adj));	/* Keep adj around */
	pw->contrast_scale = w = gtk_hscale_new(adj);
	gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	gtk_scale_set_value_pos(GTK_SCALE(w), GTK_POS_RIGHT);
	gtk_scale_set_digits(GTK_SCALE(w), 2);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(PrintDlgAdjustmentValueChangedCB), pw
	);
	GUISetWidgetTip(w, PRINT_TTM_CONTRAST);
	gtk_widget_show(w);


	/* Hbox for greyscale and color radios */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Black & White Radio */
	gslist = NULL;
	pw->visual_bw_radio = w =
	    gtk_radio_button_new_with_label(gslist, "B&W");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgToggledCB), pw
	);
	gtk_widget_show(w);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	/* Greyscale Radio */
	pw->visual_greyscale_radio = w =
	    gtk_radio_button_new_with_label(gslist, "GreyScale");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgToggledCB), pw
	);
	gtk_widget_show(w);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	/* Color Radio */
	pw->visual_color_radio = w =
	    gtk_radio_button_new_with_label(gslist, "Color");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgToggledCB), pw
	);
	gtk_widget_show(w);


	/* Orientation */
	/* Frame */
	w = gtk_frame_new("Orientation");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;
	/* Hbox */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;
	/* Portrait radio */
	gslist = NULL;
	pw->orientation_portrait_radio = w =
	    gtk_radio_button_new_with_label(gslist, "Portrait");
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgToggledCB), pw
	);
	gtk_widget_show(w);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	/* Landscape radio */
	pw->orientation_landscape_radio = w =
	    gtk_radio_button_new_with_label(gslist, "Landscape");
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgToggledCB), pw
	);
	gtk_widget_show(w);

	/* Paper size */
	w = gtk_frame_new("Paper Size");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;
	/* Vbox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;
	/* Scrolled window for clist */
	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
	    GTK_SCROLLED_WINDOW(w),
	    GTK_POLICY_AUTOMATIC,
	    GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	scroll_parent = w;
	/* Clist */
	heading[0] = "Type";
	heading[1] = "Size";
	pw->paper_size_clist = w = gtk_clist_new_with_titles(2, heading);
	clist = GTK_CLIST(w);
	gtk_widget_set_usize(w, 250, -1);
	gtk_container_add(GTK_CONTAINER(scroll_parent), w);
	gtk_widget_realize(w);
	gtk_clist_column_titles_hide(clist);
	gtk_clist_column_titles_passive(clist);
	gtk_clist_set_selection_mode(clist, GTK_SELECTION_SINGLE);
	gtk_clist_set_row_height(clist, PRINT_CLIST_ROW_SPACING);
	gtk_clist_set_column_justification(
	    clist, 0, GTK_JUSTIFY_LEFT
	);
	gtk_clist_set_column_justification(
	    clist, 1, GTK_JUSTIFY_LEFT
	);
	gtk_clist_set_column_width(clist, 0, 90);
	gtk_clist_set_column_width(clist, 1, 10);
	gtk_clist_set_shadow_type(clist, GTK_SHADOW_IN);
	gtk_signal_connect(
	    GTK_OBJECT(w), "select_row",
	    GTK_SIGNAL_FUNC(PrintDlgSelectRowCB), pw
	);
	gtk_widget_show(w);
	if(clist->columns >= 2)
	{
	    gint row;
	    paper_size_struct paper_size[] = PAPER_SIZE_LIST;
	    const paper_size_struct *src_psize = paper_size;
	    paper_size_struct *tar_psize;

	    while(src_psize->name != NULL)
	    {
		tar_psize = PAPER_SIZE(g_malloc0(sizeof(paper_size_struct)));
		if(tar_psize == NULL)
		    break;

		/* Set row cell values */
		heading[0] = src_psize->name;
		heading[1] = src_psize->desc;

		/* Append a new row to the clist */
		row = gtk_clist_append(GTK_CLIST(w), heading);
		if(row > -1)
		{
		    tar_psize->name = STRDUP(src_psize->name);
		    tar_psize->desc = STRDUP(src_psize->desc);
		    tar_psize->width = src_psize->width;
		    tar_psize->height = src_psize->height;
		    gtk_clist_set_row_data_full(
			GTK_CLIST(w), row,
			tar_psize, PrintDlgPaperSizeRowDataDeleteCB
		    );
		}
		else
		{
		    g_free(tar_psize);
		}

		src_psize++;
	    }
	}

	/* Hbox for resolution */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* DPI X */
	w = (GtkWidget *)GUIPromptBar(
	    NULL, "DPI X:",
	    NULL, &entry
	);
	pw->dpi_x_entry = (GtkWidget *)entry;
	GUISetWidgetTip((GtkWidget *)entry, PRINT_TTM_DPI_X);
	gtk_widget_set_usize((GtkWidget *)entry, 50, -1);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    (GtkObject *)entry, "changed",
	    GTK_SIGNAL_FUNC(PrintDlgDPIChangedCB), pw
	);
	gtk_widget_show(w);
	/* DPI Y */
	w = (GtkWidget *)GUIPromptBar(
	    NULL, "DPI Y:",
	    NULL, &entry
	);
	pw->dpi_y_entry = (GtkWidget *)entry;
	GUISetWidgetTip((GtkWidget *)entry, PRINT_TTM_DPI_Y);
	gtk_widget_set_usize((GtkWidget *)entry, 50, -1);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    (GtkObject *)entry, "changed",
	    GTK_SIGNAL_FUNC(PrintDlgDPIChangedCB), pw
	);
	gtk_widget_show(w);




	/* Right side vbox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;


	/* Output geometry */
	w = gtk_frame_new("Output Geometry");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;
	/* Vbox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;
	/* Hbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Output X */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_label_new("X:");
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,
	    -(gfloat)((guint16)-1 / 2),
	    (gfloat)((guint16)-1 / 2),
	    1.0f,
	    10.0f,
	    0.0f
	);
	pw->output_x_spin = w = gtk_spin_button_new(
	    adj, 1.0, 0
	);
	gtk_widget_set_usize(w, 65, -1);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), pw
	);
	gtk_widget_show(w);
	/* Output Y */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_label_new("Y:");
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,
	    -(gfloat)((guint16)-1 / 2),
	    (gfloat)((guint16)-1 / 2),
	    1.0f,
	    10.0f,
	    0.0f  
	);
	pw->output_y_spin = w = gtk_spin_button_new(
	    adj, 1.0, 0
	);
	gtk_widget_set_usize(w, 65, -1);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), pw
	);
	gtk_widget_show(w);
	/* Output Width */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_label_new("W:");
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,
	    0.0f,
	    (gfloat)((guint16)-1),
	    1.0f,
	    10.0f,
	    0.0f  
	);
	pw->output_width_spin = w = gtk_spin_button_new(
	    adj, 1.0, 0
	);
	gtk_widget_set_usize(w, 65, -1);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), pw
	);
	gtk_widget_show(w);
	/* Output Height */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_label_new("H:");
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,
	    0.0f,
	    (gfloat)((guint16)-1),
	    1.0f,
	    10.0f,
	    0.0f  
	);
	pw->output_height_spin = w = gtk_spin_button_new(
	    adj, 1.0, 0
	);
	gtk_widget_set_usize(w, 65, -1);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), pw
	);
	gtk_widget_show(w);
	/* Hbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Maintain Aspect */
	pw->output_maintain_aspect_check = w = gtk_check_button_new_with_label(
	    "Maintain Aspect"
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), pw
	);
	gtk_widget_show(w);




	/* Preview */
	w = gtk_frame_new("Preview");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;
	/* Vbox */
	w = gtk_vbox_new(TRUE, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;
	/* Hbox */
	w = gtk_hbox_new(TRUE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;
	/* Frame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
	gtk_container_border_width(GTK_CONTAINER(w), border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Nother vbox */
	w = gtk_vbox_new(TRUE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent5), w);
	gtk_widget_show(w);
	parent5 = w;
	/* Drawing area */
	pw->preview_da = w = gtk_drawing_area_new();
	/* Needs to accept focus for keyboard modifier events */
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS | GTK_CAN_DEFAULT);
	gtk_widget_add_events(
	    w,
	    GDK_EXPOSURE_MASK | GDK_ENTER_NOTIFY_MASK |
	    GDK_LEAVE_NOTIFY_MASK | GDK_KEY_PRESS_MASK |
	    GDK_KEY_RELEASE_MASK | GDK_BUTTON_PRESS_MASK |
	    GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK |
	    GDK_POINTER_MOTION_HINT_MASK | GDK_STRUCTURE_MASK
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "enter_notify_event",
	    GTK_SIGNAL_FUNC(PrintDlgPreviewEventCB),
	    pw
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "leave_notify_event",
	    GTK_SIGNAL_FUNC(PrintDlgPreviewEventCB),
	    pw
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(PrintDlgPreviewEventCB),
	    pw
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(PrintDlgPreviewEventCB),
	    pw
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(PrintDlgPreviewEventCB),
	    pw
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(PrintDlgPreviewEventCB),
	    pw
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "motion_notify_event",
	    GTK_SIGNAL_FUNC(PrintDlgPreviewEventCB),
	    pw
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "expose_event",
	    GTK_SIGNAL_FUNC(PrintDlgPreviewExposeCB),
	    pw
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "configure_event",
	    GTK_SIGNAL_FUNC(PrintDlgPreviewEventCB),
	    pw
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	GUISetWidgetTip(w, PRINT_TTM_PREVIEW);
	gtk_widget_show(w);

	/* Create preview graphic contexts */
	style = gtk_widget_get_style(w);
	if(style != NULL)
	{
	    GdkGC *gc;

	    /* Use drawing area's base GC */
	    gc = style->base_gc[GTK_STATE_NORMAL];
	    if(gc != NULL)
	    {
		gdk_gc_ref(gc);
		pw->gc_base = gc;
	    }
	}
	/* Create invert GC */
	colormap = gtk_widget_get_colormap(w);
	if(colormap == NULL)
	    colormap = gdk_colormap_get_system();
	if(colormap != NULL)
	{
	    GdkColor *cf, *cb;
	    GdkGCValues gcv;
	    GdkGCValuesMask gcv_mask = (
		GDK_GC_FOREGROUND | GDK_GC_BACKGROUND | 
		GDK_GC_FUNCTION | GDK_GC_FILL | GDK_GC_LINE_WIDTH |
		GDK_GC_LINE_STYLE | GDK_GC_CAP_STYLE | GDK_GC_JOIN_STYLE
	    );
	    cf = &gcv.foreground;
	    GDK_COLOR_SET_COEFF(cf, 0.0f, 0.0f, 0.0f)
	    GDK_COLORMAP_ALLOC_COLOR(colormap, cf)
	    cb = &gcv.background;
	    GDK_COLOR_SET_COEFF(cb, 1.0f, 1.0f, 1.0f)
	    GDK_COLORMAP_ALLOC_COLOR(colormap, cb)
	    gcv.function = GDK_INVERT;
	    gcv.fill = GDK_SOLID;
	    gcv.line_width = 1;
	    gcv.line_style = GDK_LINE_SOLID;
	    gcv.cap_style = GDK_CAP_NOT_LAST;
	    gcv.join_style = GDK_JOIN_MITER;
	    pw->gc_invert = gdk_gc_new_with_values(
		window, &gcv, gcv_mask
	    );
	    GDK_COLORMAP_FREE_COLOR(colormap, cf)
	    GDK_COLORMAP_FREE_COLOR(colormap, cb)
	}


	/* Separator */
	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	/* Buttons hbox */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Print button */
	pw->print_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_print_20x20_xpm, "Print", NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(PrintDlgPrintCB), pw
	);
	gtk_accel_group_add(
	    accelgrp, GDK_p, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_p);
	gtk_widget_show(w);

	/* Cancel button */
	pw->cancel_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_cancel_20x20_xpm,
#if defined(PROG_LANGUAGE_SPANISH)
            "Cancele", 
#elif defined(PROG_LANGUAGE_FRENCH)
            "Annuler",
#elif defined(PROG_LANGUAGE_GERMAN)
            "Heben",
#elif defined(PROG_LANGUAGE_ITALIAN)
            "Annullar",
#elif defined(PROG_LANGUAGE_DUTCH)
            "Annuler",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
            "Cancela",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
            "Kanseler",
#else
            "Cancel",
#endif
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(PrintDlgCancelCB), pw
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_c, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"                             
	);
	GUIButtonLabelUnderline(w, GDK_c);
	gtk_widget_show(w);




	/* Resize the preview drawing area */
	PrintDlgResizePreview(pw);

	/* Recreate the preview image */
	PrintDlgRecreatePreview(pw);

	return(pw);
}

/*
 *	Maps the print window.
 */
void PrintDlgMap(print_dlg_struct *pw)
{
	GtkWidget *w;

	if(pw == NULL)
	    return;

	if(!pw->map_state)
	{
	    w = pw->print_btn;
	    if(w != NULL)
	    {
		gtk_widget_grab_focus(w);
		gtk_widget_grab_default(w);
	    }

	    w = pw->toplevel;
	    if(w != NULL)
		gtk_widget_show(w);

	    pw->map_state = TRUE;
	}
}

/*
 *	Unmaps the print window.
 */
void PrintDlgUnmap(print_dlg_struct *pw)
{
	GtkWidget *w;

	if(pw == NULL)
	    return;

	if(pw->map_state)
	{
	    w = pw->toplevel;
	    if(w != NULL)
		gtk_widget_hide(w);

	    pw->map_state = FALSE;
	}
}

/*
 *	Destroys the given print window.
 */
void PrintDlgDelete(print_dlg_struct *pw)
{
	if(pw == NULL)
	    return;

	pw->src_img = NULL;

	/* Delete preview images */
	gdk_flush();
	ImgViewImageDelete(pw->preview_img);
	pw->preview_img = NULL;
	ImgViewImageDelete(pw->orig_preview_img);
	pw->orig_preview_img = NULL;


	pw->freeze_count++;

	/* Begin destroying widgets */
	GTK_WIDGET_DESTROY(pw->menu);
	GTK_WIDGET_DESTROY(pw->paper_size_clist);

	GTK_WIDGET_DESTROY(pw->output_printer_radio);
	GTK_WIDGET_DESTROY(pw->output_file_radio);
	GTK_WIDGET_DESTROY(pw->output_printer_entry);
	GTK_WIDGET_DESTROY(pw->output_file_entry);
	GTK_WIDGET_DESTROY(pw->output_file_browse_btn);

	GTK_WIDGET_DESTROY(pw->brightness_scale);
	GTK_WIDGET_DESTROY(pw->contrast_scale);

	GTK_WIDGET_DESTROY(pw->visual_bw_radio);
	GTK_WIDGET_DESTROY(pw->visual_greyscale_radio);
	GTK_WIDGET_DESTROY(pw->visual_color_radio);

	GTK_WIDGET_DESTROY(pw->orientation_portrait_radio);
	GTK_WIDGET_DESTROY(pw->orientation_landscape_radio);

	GTK_WIDGET_DESTROY(pw->dpi_x_entry);
	GTK_WIDGET_DESTROY(pw->dpi_y_entry);

	GTK_WIDGET_DESTROY(pw->output_x_spin);
	GTK_WIDGET_DESTROY(pw->output_y_spin);
	GTK_WIDGET_DESTROY(pw->output_width_spin);
	GTK_WIDGET_DESTROY(pw->output_height_spin);
	GTK_WIDGET_DESTROY(pw->output_maintain_aspect_check);

	GTK_WIDGET_DESTROY(pw->preview_da);

	GTK_WIDGET_DESTROY(pw->print_btn);
	GTK_WIDGET_DESTROY(pw->cancel_btn);

	GTK_WIDGET_DESTROY(pw->toplevel);

	GTK_ACCEL_GROUP_UNREF(pw->accelgrp);

	GTK_OBJECT_UNREF(pw->brightness_adj);
	GTK_OBJECT_UNREF(pw->contrast_adj);

	GDK_CURSOR_DESTROY(pw->busy_cur);
	GDK_CURSOR_DESTROY(pw->translate_cur);
	GDK_CURSOR_DESTROY(pw->size_cur);

	GDK_GC_UNREF(pw->gc_base);
	GDK_GC_UNREF(pw->gc_invert);

	pw->freeze_count--;

	g_free(pw);
}


/*
 *	Creates a new ImgView image suitable for printing from the
 *	specified image and the print values.
 */
imgview_image_struct *PrintDlgCreateImageFromValues(
	imgview_image_struct *img,
	gint frame_num,
	const print_values_struct *v,
	const guint8 *bg_color			/* 4 bytes RGBA */
)
{
	guint vflags;
	gint i, bpp;
	print_orientation orientation = PRINT_ORIENTATION_PORTRAIT;
	gint	output_width = 0,
		output_height = 0;
	imgview_frame_struct	*src_frame,
				*tar_frame;
	imgview_image_struct	*src_img = img,
				*tar_img;

	if((src_img == NULL) || (v == NULL))
	    return(NULL);

	src_frame = ImgViewImageGetFrame(src_img, frame_num);
	if(src_frame == NULL)
	    return(NULL);
	if(src_frame->buf == NULL)
	    return(NULL);

	bpp = src_img->bpp;

	/* Get print values */
	vflags = v->flags;
	if(vflags & PRINT_VALUE_ORIENTATION)
	    orientation = v->orientation;
	if(vflags & PRINT_VALUE_OUTPUT_WIDTH)
	    output_width = v->output_width;
	if(vflags & PRINT_VALUE_OUTPUT_HEIGHT)
	    output_height = v->output_height;

	/* Output size too small? */
	if((output_width <= 0) || (output_height <= 0))
	    return(NULL);

	/* Create the target image */
	tar_img = ImgViewImageNew(
	    output_width, output_height, bpp, 0
	);
	if(tar_img == NULL)
	    return(NULL);
	i = ImgViewImageAppendFrameNew(tar_img);
	tar_frame = ImgViewImageGetFrame(tar_img, i);

	/* Create target image depending on orientation */
	if(orientation == PRINT_ORIENTATION_LANDSCAPE)
	{
	    /* Create tempory image that will be rotated (landscape) */
	    imgview_frame_struct *tmp_frame;
	    imgview_image_struct *tmp_img = ImgViewImageNew(
	        src_img->height, src_img->width, bpp, 0
	    );
	    i = ImgViewImageAppendFrameNew(tmp_img);
	    tmp_frame = ImgViewImageGetFrame(tmp_img, i);
	    if(tmp_frame != NULL)
	    {
		/* Copy/rotate source image to tempory image */
		GUIImageBufferRotateCCW90(
		    bpp,
		    src_frame->buf,
		    src_img->width, src_img->height, src_img->bpl,
		    tmp_frame->buf,
		    tmp_img->width, tmp_img->height, tmp_img->bpl
		);

		/* Copy/resize tempory image to target image */
	        GUIImageBufferResize(
		    bpp,
		    tmp_frame->buf,
		    tmp_img->width, tmp_img->height, tmp_img->bpl,
		    tar_frame->buf,
		    tar_img->width, tar_img->height, tar_img->bpl
	        );
	    }

	    /* Delete the tempory image */
	    ImgViewImageDelete(tmp_img);
	}
	else
	{
	    /* Copy/resize source image to the target image */
	    GUIImageBufferResize(
		bpp,
		src_frame->buf,
		src_img->width, src_img->height, src_img->bpl,
		tar_frame->buf,
		tar_img->width, tar_img->height, tar_img->bpl
	    );
	}

	/* Apply background */
	GUIImageBufferFlattenWithBG(
	    bpp,
	    tar_frame->buf,
	    tar_img->width, tar_img->height, tar_img->bpl,
	    bg_color
	);

	/* Modify the target image's color, brightness, and contrast */
	PrintDlgUpdatePreviewVisualV(tar_img, v);

	return(tar_img);
}



/*
 *	Queries user to print the specified ImgView image.
 *
 *	Creates a new print window and blocks output until the user
 *	responds.
 *
 *	The specified values *v will be used to initialize the print
 *	values and *v will be modified if TRUE is returned, otherwise
 *	*v should be considered garbage after this call.
 */
gboolean PrintDlgGetResponse(
	imgview_image_struct *img,		/* Image to print */
	gint frame_num,				/* Frame to print */
	print_values_struct *v,
	const guint8 *bg_color,			/* 4 bytes RGBA */
	GtkWidget *toplevel
)
{
	GtkWidget *w;
	guint vflags;
	print_dlg_struct *pw;
	static gchar lfile_name[PATH_MAX + NAME_MAX];
#define PRINT_COMMAND_MAX	((PATH_MAX + NAME_MAX) * 2) + 1024
	static gchar lprint_command[PRINT_COMMAND_MAX];


	/* Reset globals */
	got_response = FALSE;
	main_block_level = 0;

	/* Must have print values */
	if(v == NULL)
	    return(FALSE);

	/* Create new print window */
	pw = PrintDlgNew();
	if(pw == NULL)
	    return(FALSE);


	/* Begin setting values */
	vflags = v->flags;

	/* Source ImgView image that user wants to print */
	pw->src_img = img;
	pw->src_frame_num = frame_num;

	/* Background color */
	if(bg_color != NULL)
	    memcpy(&pw->bg_color, bg_color, 4);
	else
	    memset(&pw->bg_color, 0xffffffff, 4);

	/* Output image geometry */
	if(vflags & PRINT_VALUE_OUTPUT_X)
	    pw->output_x = v->output_x;
	if(vflags & PRINT_VALUE_OUTPUT_Y)
	    pw->output_y = v->output_y;
	if(vflags & PRINT_VALUE_OUTPUT_WIDTH)
	    pw->output_width = v->output_width;
	if(vflags & PRINT_VALUE_OUTPUT_HEIGHT)
	    pw->output_height = v->output_height;
	if(vflags & PRINT_VALUE_OUTPUT_MAINTAIN_ASPECT)
	    pw->output_maintain_aspect = v->output_maintain_aspect;

	/* Update output entry widgets with the new output geometry
	 * values
	 */
	PrintDlgSyncOutputWidgets(pw);

	/* Print command */
	if(vflags & PRINT_VALUE_PRINT_COMMAND)
	{
	    w = pw->output_printer_entry;
	    if((w != NULL) ? GTK_IS_ENTRY(w) : FALSE)
		gtk_entry_set_text(
		    GTK_ENTRY(w),
		    (v->print_command != NULL) ? v->print_command : ""
		);
	}
	/* File name */
	if(vflags & PRINT_VALUE_FILE_NAME)
	{
	    w = pw->output_file_entry;
	    if((w != NULL) ? GTK_IS_ENTRY(w) : FALSE)
		gtk_entry_set_text(
		    GTK_ENTRY(w),
		    (v->file_name != NULL) ? v->file_name : ""
		);
	}
	/* Print output type */
	if(vflags & PRINT_VALUE_OUTPUT_TYPE)
	{
	    if(v->output == PRINT_OUTPUT_TO_FILE)
	    {
	        w = pw->output_printer_radio;
	        if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		    GTK_TOGGLE_BUTTON(w)->active = FALSE;
		w = pw->output_file_radio;
		if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		    GTK_TOGGLE_BUTTON(w)->active = TRUE;

		w = pw->output_file_radio;
		if(w != NULL)
		    gtk_signal_emit_by_name(GTK_OBJECT(w), "toggled");
	    }
	    else
	    {
		w = pw->output_printer_radio;
		if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		    GTK_TOGGLE_BUTTON(w)->active = TRUE;
		w = pw->output_file_radio;
		if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		    GTK_TOGGLE_BUTTON(w)->active = FALSE;

		w = pw->output_printer_radio;
		if(w != NULL)
		    gtk_signal_emit_by_name(GTK_OBJECT(w), "toggled");
	    }
	}
	/* Brightness */
	if(vflags & PRINT_VALUE_BRIGHTNESS)
	{
	    GtkAdjustment *adj = pw->brightness_adj;
	    if(adj != NULL)
	    {
		adj->value = v->brightness;
		if(adj->value < adj->lower)
		    adj->value = adj->lower;
		if(adj->value > adj->upper)
		    adj->value = adj->upper;
		gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
	    }
	}
	/* Contrast */
	if(vflags & PRINT_VALUE_CONTRAST)
	{
	    GtkAdjustment *adj = pw->contrast_adj;
	    if(adj != NULL)
	    {
		adj->value = v->contrast;
		if(adj->value < adj->lower)
		    adj->value = adj->lower;
		if(adj->value > adj->upper)
		    adj->value = adj->upper;
		gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
	    }
	}
	/* Visual */
	if(vflags & PRINT_VALUE_VISUAL)
	{
	    w = pw->visual_bw_radio;
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		GTK_TOGGLE_BUTTON(w)->active = FALSE;
	    w = pw->visual_greyscale_radio;
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		GTK_TOGGLE_BUTTON(w)->active = FALSE;
	    w = pw->visual_color_radio;
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		GTK_TOGGLE_BUTTON(w)->active = FALSE;

	    switch(v->visual)
	    {
	      case PRINT_VISUAL_BW:
		w = pw->visual_bw_radio;
		break;
	      case PRINT_VISUAL_GREYSCALE:
		w = pw->visual_greyscale_radio;
		break;
	      case PRINT_VISUAL_RGB:
		w = pw->visual_color_radio;
		break;
	    }
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
	    {
		GTK_TOGGLE_BUTTON(w)->active = TRUE;
/*		gtk_signal_emit_by_name(GTK_OBJECT(w), "toggled"); */
	    }
	}
	/* Orientation */
	if(vflags & PRINT_VALUE_ORIENTATION)
	{
	    w = pw->orientation_portrait_radio;
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		GTK_TOGGLE_BUTTON(w)->active = FALSE;
	    w = pw->orientation_landscape_radio;
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		GTK_TOGGLE_BUTTON(w)->active = FALSE;

	    switch(v->orientation)
	    {
	      case PRINT_ORIENTATION_PORTRAIT:
		w = pw->orientation_portrait_radio;
		break;
	      case PRINT_ORIENTATION_LANDSCAPE:
		w = pw->orientation_landscape_radio;
		/* Need to swap the output size for the landscape
		 * orientation
		 */
		if((vflags & PRINT_VALUE_OUTPUT_WIDTH) &&
		   (vflags & PRINT_VALUE_OUTPUT_HEIGHT)
		)
		{
		    const gint i = pw->output_width;
		    pw->output_width = pw->output_height;
		    pw->output_height = i;
		}
		break;
	    }
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
	    {
		GTK_TOGGLE_BUTTON(w)->active = TRUE;
/*		gtk_signal_emit_by_name(GTK_OBJECT(w), "toggled"); */
	    }
	}
	/* DPI X */
	if(vflags & PRINT_VALUE_DPI_X)
	{
	    gchar num_str[40];
	    w = pw->dpi_x_entry;
	    if(w != NULL)
	    {
		g_snprintf(num_str, sizeof(num_str), "%i", v->dpi_x);
		gtk_entry_set_text(GTK_ENTRY(w), num_str);
	    }
	}
	/* DPI Y */
	if(vflags & PRINT_VALUE_DPI_Y)
	{
	    gchar num_str[40];
	    w = pw->dpi_y_entry;
	    if(w != NULL)
	    {
		g_snprintf(num_str, sizeof(num_str), "%i", v->dpi_y);
		gtk_entry_set_text(GTK_ENTRY(w), num_str);
	    }
	}
	/* Paper size */
	if((vflags & PRINT_VALUE_PAPER_WIDTH) &&
	   (vflags & PRINT_VALUE_PAPER_HEIGHT)
	)
	{
	    GtkCList *clist = (GtkCList *)pw->paper_size_clist;
	    if(clist != NULL)
	    {
		gint row;
		const paper_size_struct *psize;

		/* Select the row who's paper size matches the
		 * specified paper size
		 */
		for(row = 0; row < clist->rows; row++)
		{
		    psize = PAPER_SIZE(
			gtk_clist_get_row_data(clist, row)
		    );
		    if(psize == NULL)
			continue;

		    if((v->paper_width == psize->width) &&
		       (v->paper_height == psize->height)
		    )
		    {
			gtk_clist_select_row(clist, row, 0);
			break;
		    }
		}
		/* No row selected? */
		if(row >= clist->rows)
		{
		    /* select first row if possible */
		    if(clist->rows > 0)
			gtk_clist_select_row(clist, 0, 0);
		}
	    }
	}
	/* Note that selecting a paper size above with 
	 * gtk_clist_select_row() would have called the "select_row"
	 * callback and update the preview widget and preview image
	 */

	PrintDlgSyncOutputWidgets(pw);


	/* Set reference toplevel? */
	if((toplevel != NULL) && (pw->toplevel != NULL))
	{
	    if(GTK_IS_WINDOW(toplevel))
	    {
		gtk_window_set_modal(
		    GTK_WINDOW(pw->toplevel), TRUE
	        );
	        gtk_window_set_transient_for(
		    GTK_WINDOW(pw->toplevel), GTK_WINDOW(toplevel)
	        );
	    }
	}


	/* Map print window */
	PrintDlgMap(pw);

	/* Enter main loop and wait for user response */
	main_block_level++;
	gtk_main();

	/* Break out of any additional main loop levels */
	while(main_block_level > 0)
	{
	    main_block_level--;
	    gtk_main_quit();
	}


	/* Got response? */
	if(got_response)
	{
	    /* Get the values from the print dialog and update the
	     * specified print values with them
	     */

	    /* Print output type */
	    if(vflags & PRINT_VALUE_OUTPUT_TYPE)
	    {
		w = pw->output_file_radio;
		if((w != NULL) ? GTK_TOGGLE_BUTTON(w)->active : FALSE)
		    v->output = PRINT_OUTPUT_TO_FILE;
		else
		    v->output = PRINT_OUTPUT_TO_PRINTER;
	    }
	    /* Print command */
	    if(vflags & PRINT_VALUE_PRINT_COMMAND)
	    {
		w = pw->output_printer_entry;
		if(w != NULL)
		{
		    const gchar *s = gtk_entry_get_text(GTK_ENTRY(w));
		    if(s != NULL)
		    {
			strncpy(lprint_command, s, sizeof(lprint_command));
			lprint_command[sizeof(lprint_command) - 1] = '\0';
			v->print_command = lprint_command;
		    }
		}
	    }
	    /* Print to file name */
	    if(vflags & PRINT_VALUE_FILE_NAME)
	    {
		w = pw->output_file_entry;
		if(w != NULL)
		{
		    const gchar *s = gtk_entry_get_text(GTK_ENTRY(w));
		    if(s != NULL)
		    {
			strncpy(lfile_name, s, sizeof(lfile_name));
			lfile_name[sizeof(lfile_name) - 1] = '\0';
			v->file_name = lfile_name;
		    }
		}
	    }
	    /* Brightness */
	    if(vflags & PRINT_VALUE_BRIGHTNESS)
	    {
		GtkAdjustment *adj = pw->brightness_adj;
		if(adj != NULL)
		{
		    v->brightness = adj->value;
		}
	    }
	    /* Contrast */
	    if(vflags & PRINT_VALUE_CONTRAST)
	    {
		GtkAdjustment *adj = pw->contrast_adj;
		if(adj != NULL)
		{
		    v->contrast = adj->value;
		}
	    }
	    /* Visual */
	    if(vflags & PRINT_VALUE_VISUAL)
	    {
		if(GTK_TOGGLE_BUTTON_GET_ACTIVE(
		    pw->visual_bw_radio
		))
		    v->visual = PRINT_VISUAL_BW;
		else if(GTK_TOGGLE_BUTTON_GET_ACTIVE(   
		    pw->visual_greyscale_radio 
		)) 
		    v->visual = PRINT_VISUAL_GREYSCALE; 
		else if(GTK_TOGGLE_BUTTON_GET_ACTIVE(
		    pw->visual_color_radio 
		)) 
		    v->visual = PRINT_VISUAL_RGB;
	    }
	    /* Orientation */
	    if(vflags & PRINT_VALUE_ORIENTATION)
	    {
		w = pw->orientation_landscape_radio;
		if((w != NULL) ? GTK_TOGGLE_BUTTON(w)->active : FALSE)
		    v->orientation = PRINT_ORIENTATION_LANDSCAPE;
		else
		    v->orientation = PRINT_ORIENTATION_PORTRAIT;
	    }
	    /* DPI X */
	    if(vflags & PRINT_VALUE_DPI_X)
	    {
		w = pw->dpi_x_entry;
		if((w != NULL) ? GTK_IS_ENTRY(w) : FALSE)
		{
		    const gchar *s = gtk_entry_get_text(GTK_ENTRY(w));
		    if(s != NULL)
			v->dpi_x = atoi(s);
		}
	    }
	    /* DPI Y */
	    if(vflags & PRINT_VALUE_DPI_Y)
	    {
		w = pw->dpi_y_entry;
		if((w != NULL) ? GTK_IS_ENTRY(w) : 0)
		{
		    const gchar *s = gtk_entry_get_text(GTK_ENTRY(w));
		    if(s != NULL)
			v->dpi_y = atoi(s);
		}
	    }
	    /* Paper size */
	    if((vflags & PRINT_VALUE_PAPER_WIDTH) &&
	       (vflags & PRINT_VALUE_PAPER_HEIGHT)
	    )
	    {
		/* Get paper size (without DPI applied) */
		PrintDlgPaperSize(pw, &v->paper_width, &v->paper_height, FALSE);
	    }

	    /* Output image geometry */
	    if(vflags & PRINT_VALUE_OUTPUT_X)
		v->output_x = pw->output_x;
	    if(vflags & PRINT_VALUE_OUTPUT_Y)
		v->output_y = pw->output_y;
	    if(vflags & PRINT_VALUE_OUTPUT_WIDTH)
		v->output_width = pw->output_width;
	    if(vflags & PRINT_VALUE_OUTPUT_HEIGHT)
		v->output_height = pw->output_height;

	    if(vflags & PRINT_VALUE_OUTPUT_MAINTAIN_ASPECT)
		v->output_maintain_aspect = pw->output_maintain_aspect;
	}


	/* Unset reference toplevel */
	if(pw->toplevel != NULL)
	{
	    gtk_window_set_modal(
		GTK_WINDOW(pw->toplevel), FALSE
	    );
	    gtk_window_set_transient_for(
		GTK_WINDOW(pw->toplevel), NULL
	    );
	}

	/* Delete the print window */
	PrintDlgDelete(pw);

	return(got_response);
#undef PRINT_COMMAND_MAX
}

#endif
