#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "guiutils.h"
#include "resizedlg.h"
#include "config.h"

#include "images/icon_resize_20x20.xpm"
#include "images/icon_resize_32x32.xpm"
#include "images/icon_resize_48x48.xpm"
#include "images/icon_cancel_20x20.xpm"


typedef struct _Dialog			Dialog;


/*
 *	Resize Dialog:
 */
struct _Dialog {
	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gint		freeze_count;
	GtkWidget	*maintain_aspect_check,
			*width_spin,
			*height_spin,
			*width_coeff_spin,
			*height_coeff_spin;
	GtkAdjustment	*width_adj,
			*height_adj,
			*width_coeff_adj,
			*height_coeff_adj;
	gfloat		orig_width,
			orig_height;
	gfloat		orig_aspect;		/* Width / height */
};
#define DIALOG(p)	((Dialog *)(p))


static gint delete_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void resize_cb(GtkWidget *widget, gpointer data);
static void cancel_cb(GtkWidget *widget, gpointer data);
static void coeff_changed_cb(GtkWidget *widget, gpointer data);
static void size_changed_cb(GtkWidget *widget, gpointer data);

static void build_std_dialog(
	gint width, gint height, const gchar *title,
	GtkWidget *ref_toplevel,
	GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
	GtkAccelGroup **accelgrp_rtn
);

gboolean iv_resize_opts(
	GtkWidget *ref_toplevel,
	gint *width,
	gint *height,
	gboolean *maintain_aspect
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


static gboolean		got_response;


/*
 *	Resize Dialog toplevel GtkWindow "delete_event" signal
 *	callback.
 */
static gint delete_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
        got_response = FALSE;
        gtk_main_quit();
	return(TRUE);
}

/*
 *	Resize callback.
 */
static void resize_cb(GtkWidget *widget, gpointer data)
{
	got_response = TRUE;
	gtk_main_quit();
}

/*
 *	Cancel callback.
 */
static void cancel_cb(GtkWidget *widget, gpointer data)
{
	got_response = FALSE;
	gtk_main_quit();
}

/*
 *	Coefficient GtkSpinButton "changed" signal callback.
 */
static void coeff_changed_cb(GtkWidget *widget, gpointer data)
{
	gboolean maintain_aspect;
	GtkAdjustment *adj;
	Dialog *d = DIALOG(data);
	if(d == NULL)
	    return;

	if(!GTK_IS_SPIN_BUTTON(widget))
	    return;

	if(d->freeze_count > 0)
	    return;
	else
	    d->freeze_count++;

	maintain_aspect = GTK_TOGGLE_BUTTON_GET_ACTIVE(
	    d->maintain_aspect_check
	);
	adj = gtk_spin_button_get_adjustment(
	    GTK_SPIN_BUTTON(widget)
	);

	if(widget == d->width_coeff_spin)
	{
	    gfloat new_width = adj->value * d->orig_width;
	    gtk_adjustment_set_value(d->width_adj, new_width);
	    if(maintain_aspect)
	    {
		gfloat new_height = adj->value * d->orig_height;
		gtk_adjustment_set_value(d->height_adj, new_height);
		gtk_adjustment_set_value(d->height_coeff_adj, adj->value);
	    }
	}
	else if(widget == d->height_coeff_spin)
	{
	    gfloat new_height = adj->value * d->orig_height;
	    gtk_adjustment_set_value(d->height_adj, new_height);
	    if(maintain_aspect)
	    {
		const gfloat new_width = adj->value * d->orig_width;
		gtk_adjustment_set_value(d->width_adj, new_width);
		gtk_adjustment_set_value(d->width_coeff_adj, adj->value);
	    }
	}

	d->freeze_count--;
}

/*
 *	Size GtkSpinButton "changed" signal callback.
 */
static void size_changed_cb(GtkWidget *widget, gpointer data)
{
	gboolean maintain_aspect;
	GtkAdjustment *adj;
	Dialog *d = DIALOG(data);
	if(d == NULL)
	    return;

	if(!GTK_IS_SPIN_BUTTON(widget))
	    return;

	if(d->freeze_count > 0)
	    return;
	else
	    d->freeze_count++;

	maintain_aspect = GTK_TOGGLE_BUTTON_GET_ACTIVE(
	    d->maintain_aspect_check
	);
	adj = gtk_spin_button_get_adjustment(
	    GTK_SPIN_BUTTON(widget)
	);

	/* Maintain aspect? */
	if(maintain_aspect)
	{
	    /* Original aspect known? */
	    gfloat aspect = d->orig_aspect;
	    if(aspect > 0.0f)
	    {
		/* Set the other adjustment's value in order to maintain
		 * the aspect
		 */
		if(widget == d->width_spin)
		{
		    gfloat new_height = adj->value / aspect;
		    gtk_adjustment_set_value(d->height_adj, new_height);
		}
		else if(widget == d->height_spin)
		{
		    gfloat new_width = adj->value * aspect;
		    gtk_adjustment_set_value(d->width_adj, new_width);
		}
	    }
	}

	/* Update coefficient size adjustments after updating the
	 * size adjustments
	 */
	if(maintain_aspect)
	{
	    gfloat new_coeff = (d->orig_width > 0.0f) ?
		(d->width_adj->value / d->orig_width) : 0.0f;
	    gtk_adjustment_set_value(d->width_coeff_adj, new_coeff);
	    gtk_adjustment_set_value(d->height_coeff_adj, new_coeff);
	}
	else
	{
	    gtk_adjustment_set_value(
	        d->width_coeff_adj,
	        (d->orig_width > 0.0f) ?
		    (d->width_adj->value / d->orig_width) : 0.0f
	    );
	    gtk_adjustment_set_value(
	        d->height_coeff_adj,
	        (d->orig_height > 0.0f) ?
		    (d->height_adj->value / d->orig_height) : 0.0f
	    );
	}

	d->freeze_count--;
}


/*
 *	Builds a standard toplevel dialog and returns the pointers to
 *	the created widgets.
 */
static void build_std_dialog(
	gint width, gint height, const gchar *title,
	GtkWidget *ref_toplevel,
	GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
	GtkAccelGroup **accelgrp_rtn
)
{
	gint	bw = GUI_BUTTON_HLABEL_WIDTH_DEF,
		bh = GUI_BUTTON_HLABEL_HEIGHT_DEF,
		border_major = 5;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent, *parent2;
	GtkWidget *toplevel, *main_vbox;


	/* Create keyboard accelerator group */
	accelgrp = gtk_accel_group_new();
	if(accelgrp_rtn != NULL)
	    *accelgrp_rtn = accelgrp;

	/* Create toplevel */
	toplevel = parent = w = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_set_usize(w, width, height);
	gtk_window_set_title(GTK_WINDOW(w), title);
#ifdef PROG_NAME
	gtk_window_set_wmclass(
	    GTK_WINDOW(w), "dialog", PROG_NAME
	);
#endif
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_resize_48x48_xpm);
	}
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(delete_event_cb), toplevel
	);
	gtk_container_border_width(GTK_CONTAINER(w), 0);
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	if((ref_toplevel != NULL) ?
	    GTK_IS_WINDOW(GTK_OBJECT(ref_toplevel)) : FALSE
	)
	{
	    gtk_window_set_modal(GTK_WINDOW(toplevel), TRUE);
	    gtk_window_set_transient_for(
		GTK_WINDOW(toplevel), GTK_WINDOW(ref_toplevel)
	    );
	}
	if(toplevel_rtn != NULL)
	    *toplevel_rtn = toplevel;

	main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;
	if(client_vbox_rtn != NULL)
	    *client_vbox_rtn = w;

	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Buttons */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Resize Button */
	w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_resize_20x20_xpm,
	    "Resize",
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(resize_cb), toplevel
	);
	gtk_accel_group_add(
	    accelgrp, GDK_r, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_r);
	gtk_widget_show(w);

	/* Cancel Button */
	w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_cancel_20x20_xpm,
#if defined(PROG_LANGUAGE_SPANISH)
            "Cancele", 
#elif defined(PROG_LANGUAGE_FRENCH)
            "Annuler",
#elif defined(PROG_LANGUAGE_GERMAN)
            "Heben",
#elif defined(PROG_LANGUAGE_ITALIAN)
            "Annullar",
#elif defined(PROG_LANGUAGE_DUTCH)
            "Annuler",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
            "Cancela",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
            "Kanseler",
#else
            "Cancel",
#endif
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(cancel_cb), toplevel
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_c, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_c);
	gtk_widget_show(w);
}


/*
 *	Queries user for Resize options.
 *
 *	Returns TRUE on Resize or FALSE on Cancel.
 */
gboolean iv_resize_opts(
	GtkWidget *ref_toplevel,
	gint *width,
	gint *height,
	gboolean *maintain_aspect
)
{
	gint	border_major = 5,
		border_minor = 2;
	GtkAccelGroup *accelgrp = NULL;
	GtkWidget *toplevel = NULL, *client_vbox = NULL;
	gfloat aspect = 0.0f;
	Dialog *d;

	/* Reset globals */
	got_response = FALSE;

	/* Create dialog base widgets */
	build_std_dialog(
	    -1, -1,
	    "Resize",
	    ref_toplevel,
	    &toplevel, &client_vbox, &accelgrp
	);
	d = DIALOG(g_malloc0(sizeof(Dialog)));
	d->toplevel = toplevel;
	d->accelgrp = accelgrp;
	d->freeze_count = 0;
	if((width != NULL) && (height != NULL))
	{
	    if(*height > 0)
		aspect = (gfloat)(*width) / (gfloat)(*height);
	}
	d->orig_aspect = aspect;
	d->orig_width = (width != NULL) ? *width : 1.0f;
	d->orig_height = (height != NULL) ? *height : 1.0f;


	/* Create Resize Dialog widgets */
	if(client_vbox != NULL)
	{
	    gchar *s;
	    GdkWindow *window = toplevel->window;
	    GtkAdjustment *adj;
	    GtkStyle *style = gtk_widget_get_style(toplevel);
	    GtkWidget	*w,
			*parent = client_vbox,
			*parent2, *parent3, *parent4, *parent5;


	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent = w;


	    /* Left column vbox */
	    w = gtk_vbox_new(TRUE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;
	    /* Icon */
	    w = gtk_pixmap_new_from_xpm_d(
		window,
		(style != NULL) ? &style->bg[GTK_STATE_NORMAL] : NULL,
		(guint8 **)icon_resize_32x32_xpm
	    );
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);   


	    /* Right column vbox */
	    w = gtk_vbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;

	    /* Hbox for Original Size */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* Original Size Label */
	    s = g_strdup_printf(
		"Original Size: %ix%i",
		(gint)d->orig_width, (gint)d->orig_height
	    );
	    w = gtk_label_new(s);
	    g_free(s);
	    gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    /* Hbox for Maintain Aspect */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* Maintain Aspect Check Button */
	    s = g_strdup_printf(
		"Maintain Original Aspect: %.4f",
		d->orig_aspect
	    );
	    d->maintain_aspect_check = w = gtk_check_button_new_with_label(s);
	    g_free(s);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);


	    /* Hbox for Width & Width Coeff */
	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Hbox for Width */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Alignment & Label */
	    w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);
	    gtk_widget_set_usize(w, 50, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent5 = w;
	    w = gtk_label_new("Width:");
	    gtk_container_add(GTK_CONTAINER(parent5), w);
	    gtk_widget_show(w);
	    /* Spin Button */
	    d->width_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
		1.0f,
		1.0f,
		(gfloat)((guint16)-1),
		1.0f,
		10.0f,
		0.0f 
	    );
	    d->width_spin = w = gtk_spin_button_new(
		adj,
		1.0,    /* Climb rate (0.0 to 1.0) */
		0       /* Digits */
	    );
	    gtk_widget_set_usize(w, 80, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    /* Grab focus */
	    gtk_widget_grab_focus(w);
	    gtk_widget_show(w);

	    /* Hbox for Width Coeff */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Alignment & Label */
	    w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);
	    gtk_widget_set_usize(w, 40, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent5 = w;
	    w = gtk_label_new("Coeff:");
	    gtk_container_add(GTK_CONTAINER(parent5), w);
	    gtk_widget_show(w);
	    /* Spin button */
	    d->width_coeff_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
		1.0f,
		0.0f,
		(gfloat)((guint16)-1),
		0.01f,
		0.1f,
		0.0f
	    );
	    d->width_coeff_spin = w = gtk_spin_button_new(
		adj,
		1.0,    /* Climb rate (0.0 to 1.0) */
		4       /* Digits */
	    );
	    gtk_widget_set_usize(w, 80, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);


	    /* Hbox for Height & Height Coeff */
	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Hbox for Height */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Alignment & Label */
	    w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);
	    gtk_widget_set_usize(w, 50, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent5 = w;
	    w = gtk_label_new("Height:");
	    gtk_container_add(GTK_CONTAINER(parent5), w);
	    gtk_widget_show(w);
	    /* Spin Button */
	    d->height_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
		1.0f,
		1.0f,
		(gfloat)((guint16)-1),
		1.0f,
		10.0f,
		0.0f  
	    );
	    d->height_spin = w = gtk_spin_button_new(
		adj,
		1.0,    /* Climb rate (0.0 to 1.0) */
		0       /* Digits */
	    );
	    gtk_widget_set_usize(w, 80, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    /* Hbox for Height Coeff */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Alignment & Label */
	    w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);
	    gtk_widget_set_usize(w, 40, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent5 = w;
	    w = gtk_label_new("Coeff:");
	    gtk_container_add(GTK_CONTAINER(parent5), w);
	    gtk_widget_show(w);
	    /* Spin button */
	    d->height_coeff_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
		1.0f,
		0.0f,
		(gfloat)((guint16)-1),
		0.01f,
		0.1f, 
		0.0f  
	    );
	    d->height_coeff_spin = w = gtk_spin_button_new(
		adj,
		1.0,    /* Climb rate (0.0 to 1.0) */
		4       /* Digits */
	    );
	    gtk_widget_set_usize(w, 80, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    /* Connect signals for size adjustments */
	    gtk_signal_connect(
		GTK_OBJECT(d->width_spin), "changed",
		GTK_SIGNAL_FUNC(size_changed_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(d->height_spin), "changed",
		GTK_SIGNAL_FUNC(size_changed_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(d->width_coeff_spin), "changed",
		GTK_SIGNAL_FUNC(coeff_changed_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(d->height_coeff_spin), "changed",
		GTK_SIGNAL_FUNC(coeff_changed_cb), d
	    );


	    /* Set initial values */
	    w = d->maintain_aspect_check;
	    if((maintain_aspect != NULL) ? *maintain_aspect : FALSE)
		GTK_TOGGLE_BUTTON(w)->active = TRUE;

	    adj = d->width_adj;
	    if(width != NULL)
		gtk_adjustment_set_value(adj, (gfloat)(*width));

	    adj = d->height_adj;
	    if(height != NULL)
		gtk_adjustment_set_value(adj, (gfloat)(*height));
	}


	/* Map toplevel */
	gtk_widget_show_raise(toplevel);


	/* Push main loop to wait for user response */
	gtk_main();


	/* Get values from widgets if user clicked on save */
	if(got_response)
	{
	    GtkAdjustment *adj;
	    GtkWidget *w = d->maintain_aspect_check;
	    if((maintain_aspect != NULL) && (w != NULL))
		*maintain_aspect = GTK_TOGGLE_BUTTON(w)->active;

	    adj = d->width_adj;
	    if((width != NULL) && (adj != NULL))
		*width = (gint)GTK_ADJUSTMENT_GET_VALUE(adj);

	    adj = d->height_adj;
	    if((height != NULL) && (adj != NULL))  
		*height = (gint)GTK_ADJUSTMENT_GET_VALUE(adj);
	}


	/* Destroy widgets and related resources */
	gtk_window_set_modal(GTK_WINDOW(toplevel), FALSE);
	gtk_window_set_transient_for(GTK_WINDOW(toplevel), NULL);

	GTK_WIDGET_DESTROY(d->toplevel)
	GTK_ACCEL_GROUP_UNREF(d->accelgrp)
	g_free(d);

	/* Return user response */
	return(got_response);
}
